'''
TypedLeaf: a marker for typechecking leaves.

>>> int_or_bool = TypedLeaf(int, bool)
>>> validate(int_or_bool, 12)
>>> validate(int_or_bool, True)
>>> validate(int_or_bool, "12")
Traceback (most recent call last):
...
Invalid: type_error


dictify and undictify always pass through values unchanged, regardless of type:

>>> dictify(int_or_bool, True)
True
>>> dictify(int_or_bool, "Not a bool")
'Not a bool'
>>> undictify(int_or_bool, 100)
100
>>> undictify(int_or_bool, "100")
'100'

Additionally, this module provides several TypedLeaf subclasses with the types
prefilled:

>>> s = String()
>>> validate(s, "foo")
>>> validate(s, 12)
Traceback (most recent call last):
...
Invalid: type_error

>>> b = Boolean()
>>> validate(b, True)
>>> validate(b, False)
>>> validate(b, 1)
Traceback (most recent call last):
...
Invalid: type_error

>>> i = Int()
>>> validate(i, 12)
>>> validate(i, 12.1)
Traceback (most recent call last):
...
Invalid: type_error

>>> num = Number()
>>> validate(num, 1)
>>> validate(num, 1.1)
>>> validate(num, 1+1j)
Traceback (most recent call last):
...
Invalid: type_error
>>> validate(num, 'ten')
Traceback (most recent call last):
...
Invalid: type_error

>>> com = Complex()
>>> validate(com, 1)
>>> validate(com, 1.1)
>>> validate(com, 1+1j)
>>> validate(com, 'ten')
Traceback (most recent call last):
...
Invalid: type_error

'''

import numbers
import sys

if sys.version < '3': # pragma: no cover
    unicode_type = unicode
    bytes_type = str
else: # pragma: no cover
    bytes_type = bytes
    unicode_type = str
    basestring = str

from .base import Leaf, dictify, undictify, validate
from .invalid import Invalid

class TypedLeaf(Leaf):
    '''Marker for simple types, such as integers and strings.

    Parameterized by a list of allowed types.

    TypedLeaf(*types) simply means "A value of one of these types."
    '''
    def __init__(self, *types):
        if types:
            self.types = types

    types = None

# validate checks .types,

@validate.when(TypedLeaf)
def validate_tl(dispgraph, value, **kwargs):
    if not isinstance(value, dispgraph.marker.types):
        raise Invalid('type_error')

Boolean = TypedLeaf.subclass(types=(bool,), __class_name="Boolean")
String = TypedLeaf.subclass(types=(basestring,), __class_name="String")
# TODO need tests for Bytes; doctests won't work bcs they're python-3 specific
Bytes = TypedLeaf.subclass(types=(bytes_type,), __class_name="Bytes")
Int = TypedLeaf.subclass(types=(numbers.Integral,), __class_name="Int")
Number = TypedLeaf.subclass(types=(numbers.Real,), __class_name="Number")
Complex = TypedLeaf.subclass(types=(numbers.Complex,), __class_name="Complex")

# # Most serializes don't handle complex numbers by default, so we stringify them.
# @dictify.when(Complex)
# def dictify_complex(dispgraph, value, **kwargs):
#     return unicode_type(value)

# @undictify.when(Complex)
# def undictify_complex(dispgraph, value, **kwargs):
#     try:
#         return complex(value)
#     except ValueError:
#         raise Invalid("format_error", "Invalid complex: {}".format(value))


# temporary hack for JSON compatilibity
# TODO make a separate JSON-aware dfier?
import base64

@dictify.when(Bytes)
def df_bytes(dispgraph, value, **kwargs):
    '''
    >>> dictify(Bytes(), b'\\xc3\\xbf') == u'w78='
    True
    '''
    return base64.b64encode(value).decode('ascii')

@undictify.when(Bytes)
def udf_bytes(dispgraph, value, **kwargs):
    '''
    >>> undictify(Bytes(), u'w78=') == b'\\xc3\\xbf'
    True
    >>> undictify(Bytes(), 'invalid_base64\\xc3')
    Traceback (most recent call last):
        ...
    Invalid: ...
    >>> undictify(Bytes(), ['not', 'a', 'string'])
    Traceback (most recent call last):
        ...
    Invalid: ...
    '''
    try:
        return base64.b64decode(value)
    except Exception as e:
        raise Invalid("bad_value", str(e))



'''
Some tools for aggregating error messages.

These solve the problem where you wish to perform a number of subtasks,
continuing if any of the fail, and you want to aggregate all the errors
generated while doing this.

One use case for this is in form validation with human-readable output. A
simple, pythonic validation function would iterate over the fields in the form,
and would raise an exception as soon as it found any field that wasn't valid.
But when the errors are displayed to the user, you will likely want to show ALL
the errors on the form, not just the first one found.

So instead, you want to validate each field of the form regardless of whether or
not the previous fields were valid, and collect a list of all the errors that
came up, so that you can show them all to the user.

If the form has three fields called 'foo', 'bar', and 'baz', then you would want
to generate an exception with three sub-entries - one for each of these fields -
plus a list of any errors that are not specific to one field, such as violating
a constraint involving two fields' values.

The ErrorAggregator class provides tools for doing this, and the NestedException
class defines an Exception subclass well-suited to storing these errors.

'''

from contextlib import contextmanager
from collections import OrderedDict


class NestedException(Exception):
    '''Exception with sub-exceptions.

    A NestedException has two useful fields, own_errors and sub_errors.

    own_errors is a list of root errors raised during some operation.

    sub_errors is a mapping from strings to NestedExceptions, representing the
    exceptions generated by named sub-steps in the operation.

    >>> e = NestedException()
    >>> print(e)
    <no message>
    >>> print(repr(e))
    NestedException([], OrderedDict())
    >>> raise e
    Traceback (most recent call last):
        ...
    NestedException: <no message>
    >>> bool(e)
    False
    >>> 'Yes' if e else 'No'
    'No'

    >>> e = NestedException()
    >>> e.add_own(ValueError('The value is wrong',))
    >>> print(e)
    The value is wrong
    >>> print(repr(e))
    NestedException([ValueError('The value is wrong',)], OrderedDict())
    >>> raise e
    Traceback (most recent call last):
        ...
    NestedException: The value is wrong
    >>> bool(e)
    True
    >>> 'Yes' if e else 'No'
    'Yes'

    >>> e.add_own(TypeError('Why would you use this type?'))
    >>> print(e)
    The value is wrong, Why would you use this type?
    >>> print(repr(e))
    NestedException([ValueError('The value is wrong',), TypeError('Why would you use this type?',)], OrderedDict())
    >>> raise e
    Traceback (most recent call last):
        ...
    NestedException: The value is wrong, Why would you use this type?

    >>> e = NestedException()
    >>> e.add_own(ValueError('The value is wrong',))
    >>> e.add_sub("foo", ValueError("foo is also wrong"))
    >>> print(e)
    The value is wrong; foo: [foo is also wrong]
    >>> print(e[0])
    The value is wrong
    >>> print(e['foo'])
    foo is also wrong
    >>> raise e
    Traceback (most recent call last):
        ...
    NestedException: The value is wrong; foo: [foo is also wrong]

    The add_own function can also take a full NestedException, in which case it
    the new exception will be merged into this one:

    >>> e2 = NestedException()
    >>> e2.add_own(ValueError("Alternate value problem"))
    >>> e2.add_sub("foo", TypeError("The type of foo is terrible!"))
    >>> e2.add_sub("bar", BufferError("bar failed."))
    >>> e.add_own(e2)
    >>> print(e)
    The value is wrong, Alternate value problem; bar: [bar failed.], foo: [foo is also wrong, The type of foo is terrible!]

    The .own_str() helper method returns a string representation of ONLY this
    error's messages:

    >>> print(e.own_str())
    The value is wrong, Alternate value problem

    >>> print(NestedException().own_str())
    <no message>

    '''
    def __init__(self):
        self.own_errors = []
        self.sub_errors = OrderedDict()
        super(NestedException, self).__init__()

    def __nonzero__(self):
        return bool(self.own_errors) or bool(self.sub_errors)

    def __bool__(self): # pragma: no cover
        return self.__nonzero__()

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.own_errors[item]
        return self.sub_errors[item]

    def add_own(self, exc):
        '''Add an exception to this element.'''
        if isinstance(exc, NestedException):
            self.merge(exc)
        else:
            self.own_errors.append(exc)

    def add_sub(self, key, exc):
        '''Add a sub-exception.'''
        self.sub_errors.setdefault(key, type(self)()).add_own(exc)

    def merge(self, other):
        '''Merge another NestedException into this one.'''
        for own in other.own_errors:
            self.add_own(own)
        for key, sub in other.sub_errors.items():
            self.add_sub(key, sub)

    def own_str(self):
        '''String representation of our own errors but not sub errors.'''
        if len(self.own_errors) == 0:
            return "<no message>"
        return ', '.join(str(err) for err in self.own_errors)

    def __repr__(self):
        return 'NestedException({!r}, {!r})'.format(self.own_errors, self.sub_errors)

    def __str__(self):
        n_own = len(self.own_errors)
        n_sub = len(self.sub_errors)
        if n_own+n_sub == 0:
            return "<no message>"
        own = ', '.join(str(err) for err in self.own_errors)
        other = ', '.join('{0}: [{1}]'.format(key, str(err))
                          for (key, err) in sorted(self.sub_errors.items()))
        return '; '.join(x for x in [own, other] if x)


class ErrorAggregator(object):
    '''Helper class for trying several things and accumulating all errors.

    The class attribute error_type specifies the type of error used internally,
    which should be NestedException or a subclass thereof. The catch_type class
    attribute specifies which exceptions should be caught by the aggregator; by
    default all exceptions are caught.

    Basic usage:
    >>> eg = ErrorAggregator()
    >>> eg.own_error(Exception("something_wrong"))
    >>> with eg.checking():
    ...     raise ValueError("Too many prices")
    >>> with eg.checking_sub("x"):
    ...     raise Exception("x_failed")
    >>> eg.has_errors()
    True
    >>> eg.error.own_errors
    [Exception('something_wrong',), ValueError('Too many prices',)]
    >>> eg.error.sub_errors
    OrderedDict([('x', NestedException([Exception('x_failed',)], OrderedDict()))])

    If autoraise is True, then instead of aggregating it will raise an error as
    soon as one is added:

    >>> eg = ErrorAggregator(autoraise=True)
    >>> eg.own_error(Exception("something_wrong"))
    Traceback (most recent call last):
        ...
    NestedException: something_wrong
    >>> with eg.checking_sub('foo'):
    ...     raise ValueError("Everything is terrible")
    Traceback (most recent call last):
        ...
    NestedException: something_wrong; foo: [Everything is terrible]

    Otherwise, you should call .raise_if_any() to trigger the joint exception:
    >>> e = ErrorAggregator()
    >>> e.raise_if_any()
    >>> e.own_error(Exception("unmitigated_catastrophe"))
    >>> e.raise_if_any()
    Traceback (most recent call last):
        ...
    NestedException: unmitigated_catastrophe


    '''
    error_type = NestedException
    catch_type = Exception

    def __init__(self, autoraise=False):
        self.error = self.error_type()
        self.autoraise = autoraise

    def own_error(self, err):
        '''Add a self-error.'''
        self.error.add_own(err)
        if self.autoraise:
            raise self.error

    def sub_error(self, key, err):
        '''Add a child-error.'''
        self.error.add_sub(key, err)
        if self.autoraise:
            raise self.error

    @contextmanager
    def checking(self):
        '''Context manager for try-excepting tasks.'''
        try:
            yield
        except self.catch_type as e:
            self.own_error(e)

    @contextmanager
    def checking_sub(self, key):
        '''Context manager for try-excepting subtasks.'''
        try:
            yield
        except self.catch_type as e:
            self.sub_error(key, e)

    def has_errors(self):
        '''Returns true if the aggregator holds any errors.'''
        return bool(self.error)

    def raise_if_any(self):
        '''If the aggregator has any exceptions, raise a NestedException'''
        if self.has_errors():
            raise self.error




from __future__ import unicode_literals
from collections import Counter, namedtuple


def _merge_one(lists, tails):
    for list in lists:
        if not list: continue
        h = list[0]
        if not tails[h]: break
    else:
        raise ValueError("Cannot create a consistent merge order for lists.")
    i = 0
    while i < len(lists):
        list = lists[i]
        if not list:
            lists.remove(list)
            continue
        if list[0] == h:
            list.pop(0)
            if not list:
                lists.remove(list)
                continue
            tails.subtract((list[0],))
        i += 1
    return h

def _merge(lists):
    tails = Counter()
    for list in lists:
        tails.update(list[1:])
    while lists:
        yield _merge_one(lists, tails)

def merge(lists):
    '''Merge several lists, preserving relative order.

    Raises an exception if this is impossible.

    >>> merge([[1,3,5], [2,3,6], [7]])
    [1, 2, 3, 5, 6, 7]
    >>> merge([[], [9,5,1], [10,5,1], [8,5,1], []])
    [9, 10, 8, 5, 1]
    >>> merge([[1, 2], [2, 1]])
    Traceback (most recent call last):
        ...
    ValueError: Cannot create a consistent merge order for lists.
    '''
    return list(_merge([list(l) for l in lists]))

def slice_at(list, value):
    '''Get all items in a list after the first occurrence of value.

    >>> slice_at(['a', 'b', 'c', 'd'], 'b')
    ['c', 'd']

    '''
    return list[list.index(value)+1:]

# SuperMarker is like super() for dispatch targets.
# If dispatcher(val) dispatches with the keys [k1, k2, k3], then
# dispatcher(SuperMarker(k1, val)) will dispatch with the keys [k2, k3].
SuperMarker = namedtuple('SuperMarker', ['key', 'val'])

class _BaseDispatcher(object):
    '''Abstract base class for Dispatcher and DispatchSuper.

    Subclasses must provide .dispatch_mro and ._to_keys(val)
    '''
    def get_default(self):
        for dispatcher in self.dispatch_mro:
            if dispatcher._default:
                return dispatcher._default

    def _dispatch(self, key):
        for dispatcher in self.dispatch_mro:
            v = dispatcher._self_dispatch(key)
            if v is not None:
                return v
        return None

    def dispatch(self, val):
        '''Get a function for a value. Returns None if none is set.'''
        if isinstance(val, SuperMarker):
            start_key, val = val
        else:
            start_key = None
        keys = self._to_keys(val)
        if start_key:
            keys = keys[keys.index(start_key)+1:]
        for key in keys:
            fn = self._dispatch(key)
            if fn:
                return fn
        return self.get_default()

    def call(self, val, *args, **kwargs):
        '''Invoke the dispatch target on val.

        The arguments to the function will be (self, val, *args, **kwargs).
        '''
        fn = self.dispatch(val)
        if fn is None:
            raise NotImplementedError(self._to_keys(val))
        if isinstance(val, SuperMarker):
            val = val[1]
        return fn(self, val, *args, **kwargs)

    def __call__(self, val, *args, **kwargs):
        '''Synonym for self.call()'''
        return self.call(val, *args, **kwargs)

class Dispatcher(_BaseDispatcher):
    '''Dispatcher that chooses a function based on the first argument.

    A dispatcher is a collection of similar functions plus a mechanism for
    deciding which function to call based on a single value. Each dispatcher has
    a dictionary mapping keys to functions, and a keyfn that maps values to
    lists of keys. It selects a function for a value by invoking the keyfn and
    checking each key until it finds one in its dictionary.

    By default, the keyfn uses the MRO of the type of the first argument:

    >>> d = Dispatcher()
    >>> @d.when(int)
    ... def d_int(d, v):
    ...     return "{} is an integer.".format(v)
    >>> @d.when(bool)
    ... def d_bool(d, v):
    ...     return "{} is a boolean.".format(v)
    >>> d.dispatch(12) is d_int
    True
    >>> d.dispatch(True) is d_bool
    True
    >>> d.dispatch(4+3j) is None
    True


    Dispatchers are callable. When called, a dispatcher will dispatch according
    to the first argument of the call and invoke that function, passing in the
    dispatcher itself and all the call arguments:

    >>> d(12)
    '12 is an integer.'
    >>> d(False)
    'False is a boolean.'
    >>> d(3j)
    Traceback (most recent call last):
        ...
    NotImplementedError: (<class 'complex'>, <class 'object'>)


    You can provide a custom keyfn for more complex dispatch. For example,
    consider the following keyfn that returns all prefix paths of a
    '/'-delimited path:

    >>> import re
    >>> def path_prefixes(s):
    ...     return list(reversed([s[:m.start()] for m in re.finditer('$|/', s)]))
    >>> path_prefixes('foo/bar/baz')
    ['foo/bar/baz', 'foo/bar', 'foo']

    Then a dispatcher with this keyfn will check "foo/bar/baz", "foo/bar", and
    "foo" until it finds a match:

    >>> fmt_error = Dispatcher(keyfn=path_prefixes)
    >>> @fmt_error.when("bad_email")
    ... def general_email(d, v):
    ...     return "Invalid Email"
    >>> @fmt_error.when("bad_email/fmt")
    ... def no_at(d, v):
    ...     return "Invalid Email Format"
    >>> @fmt_error.when("bad_email/fmt/no_at")
    ... def no_at(d, v):
    ...     return "Email must contain an @"
    >>> fmt_error("bad_email")
    'Invalid Email'
    >>> fmt_error("bad_email/fmt/no_at")
    'Email must contain an @'
    >>> fmt_error("bad_email/fmt/bad_domain")
    'Invalid Email Format'


    The key function can be anything that returns a list of keys:

    >>> d = Dispatcher(keyfn=lambda x:range(len(x), 0, -1))
    >>> @d.when(3)
    ... def d3(d, v):
    ...     return "{} is length 3.".format(v)
    >>> @d.when(4)
    ... def d4(d, v):
    ...     return "THERE ARE FOUR LIGHTS!"
    >>> @d.default()
    ... def dn(d, v):
    ...     return "I only know three and four."
    >>> d('foo')
    'foo is length 3.'
    >>> d([1,2,3])
    '[1, 2, 3] is length 3.'
    >>> d({'a', 'b', 'c', 'd'})
    'THERE ARE FOUR LIGHTS!'
    >>> d([])
    'I only know three and four.'
    >>> d('this string has more than four characters')
    'THERE ARE FOUR LIGHTS!'

    If you dispatch on a SuperMarker value, then all keys up to the SuperMarker's
    key will be skipped, similar to the behavior of the builtin super() type:

    >>> fmt_error(SuperMarker('bad_email/fmt', 'bad_email/fmt/no_at'))
    'Invalid Email'

    Finally, each dispatcher has an optional list of parents. You can pass this
    in on initialization, or call the .sub() method of an existing dispatcher to
    create a new dispatcher with the same keyfn and with that dispatcher as its
    sole parent.

    If a dispatcher doesn't have an entry for its key, it will check its parents
    before moving on to the next key. This allows a kind of inheritance:


    >>> my_fmt_error = fmt_error.sub()
    >>> @my_fmt_error.when("bad_email/fmt/bad_domain")
    ... def bad_domain(d, v):
    ...     return "The domain is invalid"
    >>> my_fmt_error("bad_email/fmt/no_at") # calls the parent's dispatch result
    'Email must contain an @'
    >>> my_fmt_error("bad_email/fmt/bad_domain")
    'The domain is invalid'

    Multiple inheritance behaves as you would expect:

    >>> alt_fmt_error = fmt_error.sub()
    >>> @alt_fmt_error.when("bad_email/fmt/no_at")
    ... def alt_no_at(d, v):
    ...     return "You need an @"

    >>> joint_fmt_error = Dispatcher(parents=[my_fmt_error, alt_fmt_error])
    >>> joint_fmt_error("bad_email/fmt/no_at") # inherits no_at from alt_
    'You need an @'
    >>> joint_fmt_error("bad_email/fmt/bad_domain") # inherits bad_domain from my_
    'The domain is invalid'


    Note that the dispatched functions always take the dispatcher itself as the
    first argument. This allows for recursive dispatchers. For example, consider
    dispatch based on types:

    >>> handle = Dispatcher(keyfn=lambda x:type(x).__mro__)
    >>> @handle.when(object)
    ... def handle_obj(handle, obj, indent=''):
    ...     print('{}{}'.format(indent, obj))
    >>> @handle.when(list)
    ... def handle_list(handle, l, indent=''):
    ...     print('{}List:'.format(indent))
    ...     for item in l:
    ...         handle(item, indent+'  ')


    This dispatcher prints objects, splitting apart lists:

    >>> x = [1, 2.0, ['abc', 'foo', 107]]
    >>> handle(x)
    List:
      1
      2.0
      List:
        abc
        foo
        107

    We can "subclass" this handler using the .sub() method, and add custom
    handling for a specific type:

    >>> handle2 = handle.sub()
    >>> @handle2.when(int)
    ... def handle_int(handle, i, indent=''):
    ...     print('{}The number {}'.format(indent, i))

    Because the handle_list function takes the dispatcher itself as its first
    argument, this sub-dispatcher can use the same list handler while ensuring
    that it itself will be used on sub-elements:

    >>> handle2(x)
    List:
      The number 1
      2.0
      List:
        abc
        foo
        The number 107
    '''
    def __init__(self, mapping=None, default=None, keyfn=None, parents=()):
        self.mapping = mapping.copy() if mapping else {}
        self._default = default
        self.keyfn = keyfn
        if parents:
            self.dispatch_mro = (self,) + tuple(merge(p.dispatch_mro for p in parents))
            if self.keyfn is None:
                self.keyfn = parents[0].keyfn
        else:
            self.dispatch_mro = (self,)

    def _to_keys(self, val):
        '''Convert a value to a list of keys.'''
        if self.keyfn:
            return self.keyfn(val)
        return type(val).__mro__

    def register(self, keys, fn):
        '''Register fn as the target for each key in keys'''
        for key in keys:
            self.mapping[key] = fn
        return fn

    def when(self, *keys):
        '''Decorator version of .register()'''
        def decorate(fn):
            return self.register(keys, fn)
        return decorate

    def set_default(self, fn):
        '''Set the default function to call for keys that aren't defined.

        If None, then invoking a dispatcher on an unrecognized key will raise a
        NotImplementedError.
        '''
        self._default = fn

    def default(self):
        '''Decorator version of set_default'''
        def decorate(fn):
            self.set_default(fn)
            return fn
        return decorate

    def _self_dispatch(self, key):
        if key in self.mapping:
            return self.mapping[key]
        return None

    def sub(self):
        '''Create a new dispatcher with the same keyfn and self as a parent.'''
        return Dispatcher(parents=[self], keyfn=self.keyfn)

class DispatchSuper(_BaseDispatcher):
    '''
    Like super(), but for dispatchers.

    >>> import re
    >>> def path_prefixes(s):
    ...     return list(reversed([s[:m.start()] for m in re.finditer('$|/', s)]))

    >>> disp = Dispatcher(keyfn=path_prefixes)
    >>> @disp.when("recurse")
    ... def disp_recurse(d, s, val):
    ...     print(val)
    ...     if val > 0:
    ...         d(s, val-1)
    >>> disp2 = Dispatcher(parents=[disp])
    >>> @disp2.when("recurse")
    ... def disp2_recurse(d, s, val):
    ...     print("From disp2!")
    ...     DispatchSuper(disp2, d)(s, val)
    >>> disp2("recurse", 5)
    From disp2!
    5
    From disp2!
    4
    From disp2!
    3
    From disp2!
    2
    From disp2!
    1
    From disp2!
    0

    Super supports defaults as well - if the parent has none, it will fail as
    you'd expect:

    >>> @disp2.when("not_implemented_in_parent")
    ... def disp2_niip(d, s, val):
    ...     print("This will hit parent's default, which is not implemented.")
    ...     DispatchSuper(disp2, d)(s, val)
    >>> disp2("not_implemented_in_parent", 5)
    Traceback (most recent call last):
        ...
    NotImplementedError: ['not_implemented_in_parent']

    If a parent dispatcher does have a default, it gets called:

    >>> @disp.default()
    ... def disp_default(d, s, val):
    ...     print("It is implemented now!")

    >>> disp2("not_implemented_in_parent", 5)
    This will hit parent's default, which is not implemented.
    It is implemented now!

    '''
    def __init__(self, above_disp, disp):
        self.disp = disp
        self.above_disp = above_disp
        self.dispatch_mro = slice_at(self.disp.dispatch_mro, self.above_disp)

    def _to_keys(self, val):
        return self.disp._to_keys(val)

    def call(self, val, *args, **kwargs):
        # Need to override base so that we pass in self.disp instead of self
        fn = self.dispatch(val)
        if fn is None:
            raise NotImplementedError(self._to_keys(val))
        return fn(self.disp, val, *args, **kwargs)



import sys

if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str

import vertigo as vg

from .cantrips.dispatcher import Dispatcher
from .cantrips.subclass import SubclassMixin
from .dispatch_graph import DynamicDispatchGraph

class Marker(SubclassMixin):
    '''This is a placeholder type for all types used as typemarkers.'''
    def __str__(self):
        return "<{}>".format(type(self).__name__)
    def __repr__(self):
        return str(self)

class Leaf(Marker):
    '''The base for all types that don't have children.'''

class Traversable(object):
    '''Any type that knows its own typegraph.

    All subclasses MUST set cls.typegraph to a valid typegraph.
    '''

to_typegraph = Dispatcher()
'''to_typegraph converts objects into typegraphs.

A Vertigo graph is assumed to already be typegraphs and is passed through.
A Marker gets wrapped in a PlainGraphNode.
A Traversable class T is mapped to T.typegraph.

This function is used in several places where typegraphs are needed, so that
you can say e.g. List().of(Int()) instead of the considerably more verbose
List().of(vg.PlainGraphNode(Int())).

Note also that in general any type like List that expects the typegraph to have
children will implement a method called .of() to help you make typegraphs.
'''

cls_to_typegraph = Dispatcher(keyfn=lambda t:t.__mro__)
_c2t = dict()

def associate_typegraph(cls, typegraph):
    '''associate_typegraph lets you map classes to typegraphs.

    It's used by to_typegraph, and provides a convenient way to extend
    to_typegraph for non-Traversable types.

    For example, suppose you have a namedtuple class and an associated typegraph:

    >>> from collections import namedtuple
    >>> import travesty as tv
    >>> Point3 = namedtuple('Point3', ['x', 'y', 'z'])
    >>> P3TG = tv.NamedTuple(Point3).of(x=tv.Int(), y=tv.Int(), z=tv.Int())

    To work with objects of this type, you'd need to pass in the typegraph
    everywhere:

    >>> tv.undictify(P3TG, {'x':1, 'y':2, 'z':3})

    associate_typegraph lets you set this typegraph as the default if you use the
    class itself in a travesty call:

    >>> associate_typegraph(Point3, P3TG)
    >>> tv.undictify(Point3, {'x':1, 'y':2, 'z':3})
    '''
    _c2t[cls] = typegraph

@cls_to_typegraph.when(Traversable)
def traversable_to_typegraph(d, t):
    return t.typegraph

@cls_to_typegraph.default()
def default_cls_to_typegraph(d, t):
    if t in _c2t: return _c2t[t]
    raise NotImplementedError(t)

@to_typegraph.when(vg.GraphNode)
def graphnode_to_typegraph(d, node):
    return node

@to_typegraph.when(Marker)
def marker_to_typegraph(d, marker):
    return vg.PlainGraphNode(marker)

@to_typegraph.when(type)
def type_to_typegraph(d, t):
    return cls_to_typegraph(t)

class GraphDispatcher(Dispatcher):
    '''A dispatcher with a .call better for DispatchGraphs.

    Supports multiple inheritance by e.g. GraphDispatcher([parent1, parent2]).

    Note that make_dispatcher(*parents), defined below, behaves just like
    GraphDispatcher(*parents) except with some common behaviors added to parents
    by default - consider using it when you create your own dispatchers.
    '''
    def __init__(self, parents=None):
        super(GraphDispatcher, self).__init__(parents=parents)

    def sub(self, parents=()):
        '''Create a new dispatcher with the same keyfn and self as a parent.'''
        return type(self)(parents=(self,)+tuple(parents))

    def call(self, graph, *args, **kwargs):
        graph = self._mk_graph(graph)
        return graph(*args, **kwargs)

    def _mk_graph(self, graph):
        return DynamicDispatchGraph(to_typegraph(graph), self)

class Wrapper(Marker):
    '''A root for all markers that wrap other markers directly.

    All dispatchers will, by default, apply themselves to the Wrapper's .marker
    attr. This is a convenient way to augment the behavior of a marker - make a
    wrapper around it, then add behavior to a dispatcher for your wrapper. All
    other dispatchers will simply ignore the wrapper, treating it like it were
    its .marker.

    Dan notes:
    I've gone back and forth several times on whether Wrappers should actually
    have handles on their markers or just be nodes in the graph with .sub being
    the interior wrapper. I eventually decided to go this route mainly because
    it means that you use traverse to zip an object to its typegraph; if a
    Wrapper is an actual separate node that just gets skipped, then that zip is
    no longer possible.
    '''

    def __init__(self, marker):
        assert isinstance(marker, Marker), (
            "Wrapper must be applied to a Marker. To wrap a full graph, use "
            "Wrapper.wrap(graph).")
        self.marker = marker

    @classmethod
    def wrap(cls, t, **kwargs):
        typegraph = to_typegraph(t)
        value = cls(marker=typegraph.value, **kwargs)
        # The ValueOverlay is important (instead of just making a
        # PlainGraphNode) in case the wrapped typegraph is dynamic or isn't yet
        # fully constructed (e.g. a recursive SchemaObj)
        return vg.ValueOverlay(typegraph, value)

    @classmethod
    def factory(cls, *args, **kwargs):
        return lambda m: cls(m, *args, **kwargs)

    def __str__(self):
        return "<{}({})>".format(
            type(self).__name__,
            str(self.marker).strip('<>'))

unwrap = Dispatcher(keyfn=lambda x:type(x).__mro__)
'''Unwrap a marker with Wrappers on it.

When called with only one argument m, unwrap removes all wrappers from m,
returning the innermost marker.

When called with a second argument `type`, unwrap removes wrappers until it
finds one of this type, at which point it returns it. If none is find, it will
return None.

Note that travesty's GraphDispatchers will recurse into Wrappers automatically
by default, so this function is generally not needed in a GraphDispatcher.
'''
@unwrap.when(Marker)
def return_the_marker(d, marker, type=None):
    if type is None or isinstance(marker, type):
        return marker
    return None

@unwrap.when(Wrapper)
def get_wrapped_marker(d, wrapper, type=None):
    if type is not None and isinstance(wrapper, type):
        return wrapper
    return d(wrapper.marker, type)

def core_marker(marker):
    return unwrap(marker)

# Base dispatcher for all travesty-provided types. Adding new functionality to
# this dispatcher is generally a bad idea unless you know what you're doing.
# When you add a new function to this dispatcher, it will be added to
# traverse, dictify, undictify, and validate. It should therefore be very
# general - ideally the signature should be def f(dispgraph, *args, **kwargs),
# so that any other sub-dispatchers' arguments will be unaffected by it. See
# e.g. pass_through_wrapper below
base_dispatcher = GraphDispatcher()

@base_dispatcher.when(Wrapper)
def pass_through_wrapper(dispgraph, *args, **kwargs):
    '''By default, all travesty dispatchers simple pass through Wrappers.

    This allows you to add custom behavior for certain dispatchers by
    subclassing Wrapper and customizing only the dispatcher you care about.

    For example, a wrapper might have a bunch of extra validation functions;
    traverse, dictify, etc. will ignore these and continue on to the base type,
    but you'd customize validate to run the extra validation functions.
    '''
    return dispgraph.inner(*args, **kwargs)


def make_dispatcher(parents=()):
    '''Create a new graph dispatcher.

    The graph dispatcher will inherit from base_dispatcher as well as any
    parents you specify.
    '''
    return GraphDispatcher(parents = parents + (base_dispatcher,))


# The core methods provided by dfiance

# In addition to the typegraph, traverse takes the value being traversed and an
# optional keyword argument zipgraph. If zipgraph is not provided, then the
# result is a graph wrapped around the input value. If zipgraph IS provided, the
# resulting graph will have a tuple (v, zv) at each node, where v is the value
# that node would have had without a zipgraph argument and zv is the
# corresponding value from the zipgraph.
#
# In essence, the zipgraph argument allows you to zip the traversal graph with
# another graph when the other graph is structured like the *typegraph* instead
# of like the traversal graph.
traverse = make_dispatcher()
@traverse.when(Leaf)
def traverse_object(dispgraph, value, zipgraph=None, **kwargs):
    '''Default: wrap object in leaf node'''
    if zipgraph:
        return vg.PlainGraphNode((value, zipgraph.value))
    return vg.PlainGraphNode(value)

# validate functions take the graph and a value, and raise Invalid if the value
# is somehow invalid.
validate = make_dispatcher()
@validate.when(Marker)
def validate_object(dispgraph, value, **kwargs):
    pass

# dictify turns an object into a JSON-serializable structure
dictify = make_dispatcher()
# undictify is the inverse of dictify
undictify = make_dispatcher()
# Leaves are passed through by default
@dictify.when(Leaf)
@undictify.when(Leaf)
def passthrough_tl(dispgraph, value, **kwargs):
    return value




from .base import Leaf, unwrap, core_marker, GraphDispatcher, Marker
from .base import Wrapper, Traversable, to_typegraph, make_dispatcher
from .base import traverse, validate, dictify, undictify, associate_typegraph
from .datetypes import DateTime, Date, Time, TimeDelta
from .invalid import Invalid, InvalidAggregator
from .list import List
from .mapping import SchemaMapping, StrMapping, UniMapping
from .object_marker import ObjectMarker
from .optional import Optional
from .passthrough import Passthrough
from .polymorph import Polymorph
from .schema_obj import SchemaObj
from .tuple import Tuple, NamedTuple
from .typed_leaf import TypedLeaf, Boolean, String, Bytes, Int, Number, Complex
from .validated import Validated
from .validators import Validator, InRange, OneOf, RegexMatch
from .validators import AsciiString, Email, NonEmptyString

from .document import Document, DocSet

from . import document
from . import validators


__all__ = [
    'AsciiString'
    'Boolean',
    'Bytes',
    'Complex',
    'Date',
    'DateTime',
    'DocSet',
    'Document',
    'Email',
    'GraphDispatcher',
    'InRange',
    'Int',
    'Invalid',
    'InvalidAggregator',
    'Leaf',
    'List',
    'Marker',
    'NamedTuple',
    'Number',
    'NonEmptyString',
    'ObjectMarker',
    'OneOf',
    'Optional',
    'Passthrough',
    'Polymorph',
    'RegexMatch',
    'SchemaMapping',
    'SchemaObj',
    'String',
    'Time',
    'TimeDelta',
    'Traversable',
    'Tuple',
    'TypedLeaf',
    'StrMapping',
    'UniMapping',
    'Validated',
    'Validator',
    'Wrapper',
    'associate_typegraph',
    'core_marker',
    'dictify',
    'document',
    'make_dispatcher',
    'to_typegraph',
    'traverse',
    'undictify',
    'unwrap',
    'validate',
    'validators',
]



import vertigo as vg

from .base import Marker, to_typegraph
from .base import traverse, validate, dictify, undictify
from .invalid import Invalid, InvalidAggregator

class Tuple(Marker):
    '''The following docstring is ENTIRELY SUSPECT. TODO: Fix it.

    Dictifier for tuples.

    >>> from . import Complex, Int
    >>> t = Tuple.mkgraph((Int(), Complex()))
    >>> dictify(t, (2, 3j))
    (2, 3j)
    >>> undictify(t, (7, 8+2j))
    (7, (8+2j))
    >>> print(vg.ascii_tree(traverse(t, ("7", "8+2j")), sort=True))
    root: ('7', '8+2j')
      +--0: '7'
      +--1: '8+2j'

    >>> undictify(t, 7j)
    Traceback (most recent call last):
        ...
    Invalid: not_iterable
    >>> undictify(t, (1,))
    Traceback (most recent call last):
        ...
    Invalid: bad_len
    >>> validate(t, (6, 1))
    >>> validate(t, 3)
    Traceback (most recent call last):
        ...
    Invalid: not_iterable
    >>> validate(t, ("this", "tuple", "is", "too", "long"))
    Traceback (most recent call last):
        ...
    Invalid: bad_len
    >>> validate(t, (1.5, "hello"))
    Traceback (most recent call last):
        ...
    Invalid: 0: [type_error], 1: [type_error]

    It does support naming the attributes for the purposes of error reporting
    and walking the tree; see also the NamedTuple dfier.

    >>> t2 = Tuple(field_names=("num", "cnum")).of(num=Int(),cnum=Complex())
    >>> validate(t2, (1.5, "hello"))
    Traceback (most recent call last):
        ...
    Invalid: cnum: [type_error], num: [type_error]

    You must provide either a number of fields or a set of field names:

    >>> Tuple()
    Traceback (most recent call last):
        ...
    ValueError: nfields or field_names is required.
    '''
    def __init__(self, nfields=None, field_names=None):
        super(Tuple, self).__init__()
        if field_names is None:
            if nfields is None:
                raise ValueError("nfields or field_names is required.")
            field_names = [str(i) for i in range(nfields)]
        if nfields is None:
            nfields = len(field_names)
        self.field_names = field_names
        self.nfields = nfields

    def of(self, **kwargs):
        children = {key:to_typegraph(val) for key, val in kwargs.items()}
        return vg.PlainGraphNode(self, children)

    @classmethod
    def mkgraph(cls, type_tuple):
        # type_tuple should be a tuple of typegraphs
        children = {str(k):val for k, val in enumerate(type_tuple)}
        return cls(nfields=len(type_tuple)).of(**children)


@dictify.when(Tuple)
def dictify_tuple(dispgraph, value, **kw):
    names = dispgraph.marker.field_names
    return tuple(dispgraph[name](v, **kw) for name,v in zip(names, value))

@undictify.when(Tuple)
def undictify_tuple(dispgraph, value, **kw):
    names = dispgraph.marker.field_names
    try:
        if len(value) != len(names):
            raise Invalid('bad_len')
    except TypeError:
        raise Invalid('not_iterable')
    error_agg = InvalidAggregator(autoraise = kw.get('fail_early', False))
    l = []
    for (n, val) in zip(names, value):
        with error_agg.checking_sub(n):
            l.append(dispgraph[n](val, **kw))
    error_agg.raise_if_any()
    return tuple(l)

@validate.when(Tuple)
def validate_tuple(dispgraph, value, **kw):
    names = dispgraph.marker.field_names
    try:
        if len(value) != len(names):
            raise Invalid('bad_len')
    except TypeError:
        raise Invalid('not_iterable')
    error_agg = InvalidAggregator(autoraise = kw.get('fail_early', False))
    for (n, val) in zip(names, value):
        with error_agg.checking_sub(n):
            dispgraph[n](val, **kw)
    error_agg.raise_if_any()

@traverse.when(Tuple)
def traverse_tuple(dispgraph, value, zipgraph=None, **kwargs):
    def get(key):
        if hasattr(value, key):
            return getattr(value, key)
        try:
            key = int(key)
        except:
            raise AttributeError("'{}' object has no attribute '{}'".format(type(value), key))
        return value[key]
    edges = []
    names = dispgraph.marker.field_names
    for key in names:
        subgraph = dispgraph[key]
        subzip = zipgraph[key] if zipgraph else None
        edges.append((key, subgraph(get(key), subzip, **kwargs)))
    v = value
    if zipgraph:
        v = (v, zipgraph.value)
    return vg.PlainGraphNode(v, edges)



class NamedTuple(Tuple):
    '''
    Like Tuple, but for namedtuple instances:

    >>> from collections import namedtuple
    >>> from . import Complex, Int
    >>> WeirdPoint = namedtuple("WeirdPoint", ["num", "cnum"])
    >>> t = NamedTuple(WeirdPoint).of(num=Int(), cnum=Complex())
    >>> p = WeirdPoint(3, 6j)
    >>> dictify(t, p)
    (3, 6j)
    >>> undictify(t, (5, 2+1.0j))
    WeirdPoint(num=5, cnum=(2+1j))
    >>> validate(t, p)
    >>> g = vg.from_dict(dict(num="A Number", cnum="A Complex Number"))
    >>> print(vg.ascii_tree(traverse(t, p, zipgraph=g)))
    root: (WeirdPoint(num=3, cnum=6j), None)
      +--num: (3, 'A Number')
      +--cnum: (6j, 'A Complex Number')

    Note that validate explicitly expects the particular namedtuple class, and
    won't settle even for an identical one:

    >>> WeirdPoint2 = namedtuple("WeirdPoint", ["num", "cnum"])
    >>> validate(t, WeirdPoint2(3, 6j))
    Traceback (most recent call last):
        ...
    Invalid: type_error
    '''
    def __init__(self, tuple_type):
        super(NamedTuple, self).__init__(field_names=tuple_type._fields)
        self.tuple_type = tuple_type

@validate.when(NamedTuple)
def validate_namedtuple(dispgraph, value, **kw):
    tuple_type = dispgraph.marker.tuple_type
    if not isinstance(value, tuple_type):
        raise Invalid('type_error')
    dispgraph.super(NamedTuple)(value, **kw)


@undictify.when(NamedTuple)
def undictify_namedtuple(dispgraph, value, **kw):
    t = dispgraph.super(NamedTuple)(value, **kw)
    return dispgraph.marker.tuple_type._make(t)



# -*- coding: utf-8 -*-
import re
import sys

if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str

from .base import Marker, validate
from .typed_leaf import String
from .invalid import Invalid
from .validated import Validated
from .inference import infer_marker

class Validator(Marker):
    '''Validators are plug-ins to add advanced validation.

    They define a .validate(value, **kwargs) function, which raises an Invalid
    if the value is invalid and otherwise does nothing.

    You can pass a list of validators to the Validated wrapper type to apply
    advanced validation.

    Validator is actually a Marker subclass, so you can use it directly in a
    typegraph, but in general there should be no reason to do so - prefer to
    use it in conjunction with Validated.
    '''
    def validate(self, value, **kwargs): # pragma: no cover
        raise NotImplementedError()

    def of(self, marker):
        return Validated(marker, [self])

@validate.when(Validator)
def validate_validator(dispgraph, value, **kwargs):
    dispgraph.marker.validate(value, **kwargs)

class IsInRange(Validator):
    '''Require values to be between two endpoints.

    The endpoints themselves are allowed.

    >>> RangeInt = IsInRange(0, 5)
    >>> RangeInt.validate(0)
    >>> RangeInt.validate(3)
    >>> RangeInt.validate(5)
    >>> RangeInt.validate(None)
    >>> RangeInt.validate(-1)
    Traceback (most recent call last):
        ...
    Invalid: range_error/too_low
    >>> RangeInt.validate(6)
    Traceback (most recent call last):
        ...
    Invalid: range_error/too_high

    Note that the types can be any type for which < and > are defined:

    >>> from datetime import date
    >>> DateIn2012 = IsInRange(date(2012, 1, 1), date(2012, 12, 31))
    >>> DateIn2012.validate(date(2012, 4, 13))
    >>> DateIn2012.validate(date(2013, 1, 1))
    Traceback (most recent call last):
        ...
    Invalid: range_error/too_high
    '''
    def __init__(self, low=None, high=None):
        self.low = low
        self.high = high

    def validate(self, value, **kwargs):
        if value is None: return
        if self.low is not None and value < self.low:
            raise Invalid("range_error/too_low")
        if self.high is not None and value > self.high:
            raise Invalid("range_error/too_high")

def InRange(low=None, high=None, marker=None):
    '''Require values to be between two endpoints.

    See IsInRange.

    >>> SmallInt = InRange(0, 5)
    >>> SmallInt
    <Validated(Int)>
    >>> EarlyStr = InRange('a', 'h')
    >>> EarlyStr
    <Validated(String)>
    >>> validate(EarlyStr, "azkaban")
    >>> validate(EarlyStr, "zazzerpan")
    Traceback (most recent call last):
        ...
    Invalid: range_error/too_high

    You can override the inferred marker:

    >>> from . import Bytes
    >>> EarlyBytes = InRange(b'A', b'C', marker=Bytes())
    >>> validate(EarlyBytes, b'Angela McAscii')
    >>> validate(EarlyBytes, u'Alric von Üñîçø∂é')
    Traceback (most recent call last):
        ...
    Invalid: type_error
    '''

    if (marker or low or high) is None:
        raise ValueError("low, high, and marker cannot all be None")
    if not marker:
        val = low if low is not None else high
        marker = infer_marker(val)
    return IsInRange(low, high).of(marker)

class IsOneOf(Validator):
    '''Require values to come from a fixed list of options.

    >>> Baggins = IsOneOf(['Frodo', 'Bilbo'])
    >>> Baggins.validate("Bilbo")
    >>> Baggins.validate("Frodo")
    >>> Baggins.validate("Meriadoc")
    Traceback (most recent call last):
        ...
    Invalid: invalid_choice
    '''
    def __init__(self, options):
        self.options = options

    def validate(self, value, **kwargs):
        if value not in self.options:
            raise Invalid("invalid_choice")

# This would take (*options, marker=None) but python2 doesn't allow that
def OneOf(*options, **kwargs):
    '''Require values to come from a fixed list of options.

    This creates a Validated instance with an inferred marker and an IsOneOf
    validator:

    >>> Baggins = OneOf('Frodo', 'Bilbo')
    >>> Baggins
    <Validated(String)>
    >>> Baggins.vdators
    (<IsOneOf>,)
    >>> OneToThree = OneOf(1,2,3)
    >>> OneToThree
    <Validated(Int)>

    Note that the inference is based on the first argument, which may not do
    what you expected with e.g. numeric types:

    >>> from . import Complex
    >>> FourthRootOfUnity = OneOf(1, 1j, -1, -1j)
    >>> FourthRootOfUnity # You might expect a Complex
    <Validated(Int)>

    You can override the inferred marker to get around this:

    >>> FourthRootOfUnity = OneOf(1, 1j, -1, -1j, marker=Complex())
    >>> FourthRootOfUnity
    <Validated(Complex)>

    Note that you must provide a marker if you're creating an empty option
    list for some reason:

    >>> OneOf()
    Traceback (most recent call last):
        ...
    ValueError: Must provide marker or at least one option
    '''

    marker = kwargs.pop("marker", None)
    if kwargs:
        msg = "OneOf() got an unexpected keyword argument '{}'"
        key = next(iter(kwargs.keys()))
        raise TypeError(msg.format(key))
    if marker is None:
        if not options:
            raise ValueError("Must provide marker or at least one option")
        marker = infer_marker(options[0])
    return IsOneOf(options).of(marker)

class MatchesRegex(Validator):
    '''Require non-empty values to match a regular expression.

    >>> IP = MatchesRegex("\d{1,3}(\.\d{1,3}){3}")
    >>> IP.validate("127.0.0.1")
    >>> IP.validate("This is not an IP address.")
    Traceback (most recent call last):
        ...
    Invalid: invalid_string
    >>> IP.validate("127.0.0.no")
    Traceback (most recent call last):
        ...
    Invalid: invalid_string
    >>> IP.validate(12)
    Traceback (most recent call last):
        ...
    Invalid: type_error
    '''
    def __init__(self, regex):
        self.regex = re.compile(regex)

    def validate(self, value, **kwargs):
        if not isinstance(value, basestring):
            raise Invalid("type_error")
        if not self.regex.match(value):
            raise Invalid("invalid_string")

def RegexMatch(regex, marker=String()):
    return MatchesRegex(regex).of(marker)

class IsAsciiString(Validator):
    def validate(self, value, **kwargs):
        if not isinstance(value, basestring):
            raise Invalid("type_error")
        try:
            value.encode('ascii')
        except Exception:
            raise Invalid("not_ascii")

def AsciiString(marker=String()):
    return IsAsciiString().of(marker)

class IsEmail(IsAsciiString):
    '''Require a valid email address.

    Doesn't support weirder email addresses, like those with parentheses.

    >>> email = IsEmail()
    >>> email.validate('dplepage@gmail.com')
    >>> email.validate('madeupname@nowhere.none')
    >>> email.validate('foo AT bar DOT com')
    Traceback (most recent call last):
        ...
    Invalid: invalid_email/no_at
    >>> email.validate('()@bar.com')
    Traceback (most recent call last):
        ...
    Invalid: invalid_email/bad_user
    >>> email.validate('foo@')
    Traceback (most recent call last):
        ...
    Invalid: invalid_email/bad_domain
    >>> email.validate(u'føö@bår.côm')
    Traceback (most recent call last):
        ...
    Invalid: not_ascii
    >>> email.validate(12)
    Traceback (most recent call last):
        ...
    Invalid: type_error

    '''
    # regexes borrowed from formencode.validators
    usernameRE = re.compile(r"^[^ \t\n\r@<>()]+$", re.I)
    domainRE = re.compile(r'''
        ^(?:[a-z0-9][a-z0-9\-]{0,62}\.)+ # (sub)domain - alpha followed by 62max chars (63 total)
        [a-z]{2,}$                       # TLD
    ''', re.I | re.VERBOSE)

    def validate(self, value, **kwargs):
        super(IsEmail, self).validate(value, **kwargs)
        if '@' not in value:
            raise Invalid("invalid_email/no_at")
        user, domain = value.split('@', 1)
        if not self.usernameRE.match(user):
            raise Invalid("invalid_email/bad_user")
        if not self.domainRE.match(domain):
            raise Invalid("invalid_email/bad_domain")

def Email(marker=String()):
    return IsEmail().of(marker)

class IsNonEmptyString(Validator):
    '''Require something more than an empty string.

    >>> not_empty = IsNonEmptyString()
    >>> not_empty.validate(' ')
    Traceback (most recent call last):
        ...
    Invalid: empty
    '''
    def validate(self, value, **kwargs):
        if not isinstance(value, basestring):
            raise Invalid("type_error")
        if len(value.strip()) == 0:
            raise Invalid("empty")

def NonEmptyString(marker=String()):
    return IsNonEmptyString().of(marker)

if __name__ == '__main__': # pragma: no cover
    import doctest
    doctest.testmod()



import sys

if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str

import vertigo as vg

from .cantrips.error_aggregator import ErrorAggregator, NestedException

class SingleInvalid(Exception):
    '''A leaf in the nested tree of Invalid exceptions.

    In general, don't raise SingleInvalid directly - instead, raise Invalid

    self.err_id will be a string indicating the specific error, e.g. 'required'
    when a value is missing that shouldn't be, 'not_int' for a value that must
    be an integer but cannot be parsed as one, and so forth. The specific errors
    raised by various Dictifiers are class-specific.

    self.desc is an optional human-readable description of the error. This is
    useful for debugging and for simple applications; for more complex
    applications the best way to generate human-readable strings is via an
    external mapping from err_ids to descriptive strings.

    '''
    def __init__(self, err_id=None, desc=None, **kwargs):
        if kwargs:
            super(SingleInvalid, self).__init__(err_id, kwargs)
        else:
            super(SingleInvalid, self).__init__(err_id)
        self.desc = desc
        self.err_id = err_id
        self.kwargs = kwargs

    def __str__(self):
        bits = [self.err_id, self.desc]
        prefix = u' - '.join(unicode(m) for m in bits if m)
        if self.kwargs:
            return prefix + ' - {}'.format(self.kwargs)
        return prefix

    def id_string(self):
        return self.err_id


class Invalid(NestedException):
    '''Base class for all undictification errors.

    This is a NestedException where the own_errors are all SingleInvalids.
    '''
    def __init__(self, err_id=None, desc=None, **kwargs):
        super(Invalid, self).__init__()
        if err_id or desc or kwargs:
            self.own_errors.append(SingleInvalid(err_id, desc, **kwargs))

    def as_graph(self):
        edges = {k:v.as_graph() for (k,v) in self.sub_errors.items()}
        return vg.PlainGraphNode(self.own_errors, edges)

    def id_string(self):
        n_own = len(self.own_errors)
        n_sub = len(self.sub_errors)
        if n_own+n_sub == 0:
            return "<no message>"
        own = ', '.join(err.id_string() for err in self.own_errors)
        other = ', '.join('{0}: [{1}]'.format(key, err.id_string())
                          for (key, err) in sorted(self.sub_errors.items()))
        return '; '.join(x for x in [own, other] if x)


class InvalidAggregator(ErrorAggregator):
    '''Specialized ErrorAggregator that only aggregates Invalids'''
    error_type = Invalid
    catch_type = Invalid



import functools

def has_cache(fn):
    '''Add a cache dictionary to a function.

    The function must accept a keyword-only argument named 'cached'. If cache is
    provided when calling this function, it will be used normally; if it is
    omitted, then the function's own ._cache attribute, which is created by this
    decorator, will be used.

    This decorator does NOT automatically cache anything - it is up to the
    function to decide what to do with its cache.

    For example:

    >>> @has_cache
    ... def foo(x, cache):
    ...     if x not in cache:
    ...         print("Populating {}".format(x))
    ...         cache[x] = x+1
    ...     return cache[x]
    >>> foo(12)
    Populating 12
    13
    >>> foo(12)
    13
    >>> foo._cache
    {12: 13}
    >>> foo(0)
    Populating 0
    1
    >>> foo(0)
    1
    >>> foo(0, cache={})
    Populating 0
    1

    '''
    fn._cache = {}
    @functools.wraps(fn)
    def newfn(*args, **kwargs):
        if 'cache' not in kwargs:
            kwargs['cache'] = fn._cache
        return fn(*args, **kwargs)
    return newfn



'''

This is a hastily-made set of examples showing travesty, and especially
highlighting what it can do that dfiance can't.

Also it's worth noting that anything in here marked TODO might, in fact, have
already been done - don't assume that TODO means it doesn't exist (but it still
might not)

'''
import datetime
import textwrap

import travesty as tv
from travesty import traverse, validate, dictify, undictify

import vertigo as vg

# TODO reversed=True should be the default for vg.overlay (it makes more sense)
def overlay(*graphs):
    return vg.overlay(*graphs, reversed=True)

# TODO add this helper to vertigo - it's useful
def flat_overlay(graph, flat_dict):
    return overlay(graph, vg.from_flat(flat_dict))



'''
Here's an example using SchemaObj.

It's pretty similar to the one in README2.rst.

'''

class BlogPost(tv.SchemaObj):
    field_types = dict(author=tv.String(), text=tv.String(), timestamp=tv.DateTime())
    def __init__(self, author, text, timestamp=None):
        self.author = author
        self.text = text
        self.timestamp = timestamp or datetime.datetime.now()

    def __repr__(self):
        return "Post by {} at {}".format(self.author, self.timestamp)

class Blog(tv.SchemaObj):
    field_types = dict(
        title = tv.String(),
        posts = tv.List().of(BlogPost))
    def __init__(self, title, posts=()):
        self.title = title
        self.posts = list(posts)

    def __repr__(self):
        return "'{}' - {} posts".format(self.title, len(self.posts))

# 00:00 on Jan 15, 2014
start = datetime.datetime(2014, 1, 15, 00, 00)
# Generate a datetime for an hour and a day after the above start
def _time(day, hr):
    return start + datetime.timedelta(days=day, hours=hr)

# Here's an example blog
blog = Blog("The TTB Blog", posts=[
    BlogPost("dplepage", "This is my first post!", _time(0,13)),
    BlogPost("dplepage", "Lorem ipsum, and so forth and so on.", _time(1,12)),
    BlogPost("bdarklighter", "Meh.", _time(2,14)),
])

# Dictify it!
blog_dict = dictify(Blog, blog)
# we get a nice dict
assert blog_dict == {
    'posts': [
        {
            'author': 'dplepage',
            'text': 'This is my first post!',
            'timestamp': u'2014-01-15T13:00:00',
        },{
            'author': 'dplepage',
            'text': 'Lorem ipsum, and so forth and so on.',
            'timestamp': u'2014-01-16T12:00:00',
        },{
            'author': 'bdarklighter',
            'text': 'Meh.',
            'timestamp': u'2014-01-17T14:00:00',
        }],
    'title': 'The TTB Blog'
}, str(blog_dict)

# undictify gets us our blog back
b2 = undictify(Blog, blog_dict)

assert b2.title == blog.title
assert all(p1.text == p2.text for (p1, p2) in zip(b2.posts,blog.posts))


# traverse wraps the whole thing in a graph
assert vg.ascii_tree(traverse(Blog, blog), sort=True) == '''
root: 'The TTB Blog' - 3 posts
  +--posts: [Post by dplepage at 2014-01-15 13:00:00, Post by dplepage at 2014-01-16 12:00:00, Post by bdarklighter at 2014-01-17 14:00:00]
  |  +--0: Post by dplepage at 2014-01-15 13:00:00
  |  |  +--author: 'dplepage'
  |  |  +--text: 'This is my first post!'
  |  |  +--timestamp: datetime.datetime(2014, 1, 15, 13, 0)
  |  +--1: Post by dplepage at 2014-01-16 12:00:00
  |  |  +--author: 'dplepage'
  |  |  +--text: 'Lorem ipsum, and so forth and so on.'
  |  |  +--timestamp: datetime.datetime(2014, 1, 16, 12, 0)
  |  +--2: Post by bdarklighter at 2014-01-17 14:00:00
  |     +--author: 'bdarklighter'
  |     +--text: 'Meh.'
  |     +--timestamp: datetime.datetime(2014, 1, 17, 14, 0)
  +--title: 'The TTB Blog'
'''.strip()


'''
Typegraph manipulation

Let's make a validation graph that confirms that posts aren't too short.

One way to do this is to actually change the definition of BlogPost to impose
limits on its string, but this means that BlogPost will always have this
restriction, which is not ideal. I recommend that for field_types you try to
stick to just documenting the actual types of things, not validation constraints
that may not always hold.
'''

class LimitedLength(tv.Validator):
    def __init__(self, minlen=None, maxlen=None):
        self.maxlen = maxlen
        self.minlen = minlen

    def validate(self, value, **kwargs):
        if self.maxlen is not None and len(value) > self.maxlen:
            raise tv.Invalid("length/too_long", "Exceeded maximum length of {}".format(self.maxlen))
        if self.minlen is not None and len(value) < self.minlen:
            raise tv.Invalid("length/too_short", "Shorter than minimum length of {}".format(self.minlen))

# Create a new graph that looks just like the blog typegraph but replacing the
# value at posts/sub/text with a Validated wrapper.

g = flat_overlay(Blog.typegraph, {
        'posts/sub/text': tv.Validated(
            Blog.typegraph['posts', 'sub', 'text'].value,
            vdators = [LimitedLength(minlen=5, maxlen=30)],
        )
    }
)

try:
    validate(g, blog)
except tv.Invalid as e:
    # Two of our three posts are too short according to this validation graph:
    tree1 = vg.ascii_tree(e.as_graph(), sort=True)
    tree2 = textwrap.dedent('''
    root: []
      +--posts: []
         +--1: []
         |  +--text: [SingleInvalid('length/too_long',)]
         +--2: []
            +--text: [SingleInvalid('length/too_short',)]
    ''').strip()
    assert tree1 == tree2, tree1
else: # pragma: no cover
    raise RuntimeError("That should have failed.")







'''
Here's a nested type for a calendar. It's a dict with one key, 'events', which
is a list of events. Each event is a dict of (start, end, label, attendees),
where start and end are datetimes, label is a string, and attendees is a list of
strings.

I'm assembling it here with PlainGraphNode.build, so the '_self' keys indicate
values for nodes and all other keys are edges in the graph.

TODO: make some easier ways to write these things out.
'''

calendar_type = vg.PlainGraphNode.build(dict(
    _self = tv.SchemaMapping(),
    events = dict(
        _self = tv.List(), # a calendar is a list of events
        sub = dict(
            _self = tv.SchemaMapping(),
            start = tv.DateTime(),
            end = tv.DateTime(),
            label = tv.String(),
            attendees = dict(
                _self = tv.List(),
                sub = tv.String(),
            ),
        ),
    ),
))

# Update: I added some cooler ways to make these structures. Here's a less
# verbose construction:

calendar_type = tv.SchemaMapping().of(
    events = tv.List().of(
        tv.SchemaMapping().of(
            start = tv.DateTime(),
            end = tv.DateTime(),
            label = tv.String(),
            attendees = tv.List().of(tv.String()),
        ),
    ),
)



'''Here's a small calendar instance.'''

# 00:00 on Jan 15, 2014
start = datetime.datetime(2014, 1, 15, 00, 00)
# Generate a datetime for an hour and a day after the above start
def _time(day, hr):
    return start + datetime.timedelta(days=day, hours=hr)

calendar = dict(events=[
    dict(
        start=_time(0, 14),
        end=_time(0, 15),
        label = "The TTP Project - group meeting",
        attendees = ['Dan', 'Kai'],
    ),
    dict(
        start=_time(1, 11),
        end=_time(1, 12),
        label = "Attack Death Star",
        attendees = ['Biggs Darklighter', 'Wedge Antilles', 'Luke Skywalker', 'Wilhuff Tarkin'],
    ),
    dict(
        start=_time(1, 14),
        end=_time(1, 15),
        label = "Party on Endor",
        attendees = ['Wedge Antilles', 'Luke Skywalker'],
    ),
])



# dictify and undictify take the typegraph as their first args
c_struct = dictify(calendar_type, calendar)
cal2 = undictify(calendar_type, c_struct)
assert cal2 == calendar
# Note that DateTime defaults to a string
assert c_struct['events'][0]['start'] == '2014-01-15T14:00:00'


'''

Now suppose instead of strings we want to output datetimes as dicts with keys
'year', 'month', 'day', 'hour', 'minute', 'second', because we're going to send
our objects to some system that can't parse ISO dates.

There are two different ways to do this: you can change the typegraph to put
your own custom type in, or you use custom dictify/undictify functions to handle
datetimes differently.

Either way, we'll need these two functions:
'''

def datetime_to_dict(value):
    return dict(
        year = value.year,
        month = value.month,
        day = value.day,
        hour = value.hour,
        minute = value.minute,
        second = value.second,
    )

def dict_to_datetime(value):
    return datetime.datetime(
        year = value['year'],
        month = value['month'],
        day = value['day'],
        hour = value['hour'],
        minute = value['minute'],
        second = value['second'],
    )

# Option one: custom type replacement
class DictDateTime(tv.DateTime):
    pass

@dictify.when(DictDateTime)
def dfy_dt(dispgraph, value, **kwargs):
    return datetime_to_dict(value)

@undictify.when(DictDateTime)
def undfy_dt(dispgraph, value, **kwargs):
    return dict_to_datetime(value)

# Create a new typegraph by overlaying a sparse graph on calendar_type
ct2 = flat_overlay(calendar_type, {
    'events/sub/start':DictDateTime(),
    'events/sub/end':DictDateTime(),
})

# The new type markers cause dictify/undictify to use the functions above
c_struct = dictify(ct2, calendar)
cal2 = undictify(ct2, c_struct)
assert cal2 == calendar
# Now the serialized form has a dict
assert c_struct['events'][0]['start'] == dict(
    year=2014,
    month=1,
    day=15,
    hour=14,
    minute=0,
    second=0)



# Option two: custom dispatcher

df2 = dictify.sub()
udf2 = undictify.sub()

# These functions look the same as before, but now we're telling our new dictify
# and undictify functions to handle ALL DateTimes differently.
@df2.when(tv.DateTime)
def dfy_dt2(dispgraph, value, **kwargs):
    return datetime_to_dict(value)

@udf2.when(tv.DateTime)
def undfy_dt2(dispgraph, value, **kwargs):
    return dict_to_datetime(value)

# This has the same behavior as the first case, but this time we didn't have to
# modify the typegraph.

c_struct = df2(calendar_type, calendar)
cal2 = udf2(calendar_type, c_struct)
assert cal2 == calendar
assert c_struct['events'][0]['start'] == dict(
    year=2014,
    month=1,
    day=15,
    hour=14,
    minute=0,
    second=0)





'''
Option 1 is more flexible than 2, because you can e.g. choose to have some dates
dictified one way and some dictified another.

Option 2 is powerful when you want to make sure that you hit ALL dates, instead
of picking and choosing, maybe because picking and choosing is too much work, or
because you're making a general function that could be applied to any typegraph.

'''






import vertigo as vg

from .cantrips.empty_instance import create_instance

from .base import traverse, validate, dictify, undictify, to_typegraph
from .invalid import Invalid, InvalidAggregator
from .schema import Schema

class ObjectMarker(Schema):
    '''Marker for objects that can be assembled by field.

    The Marker is parameterized by a target_cls, which determines what type is
    expected.

    The children of an ObjectMarker node in the typegraph identify the salient
    attributes of the target_cls; dictify will use getattr to extract and store
    those attributes, and undictify will unpack the stored attributes and pass
    them into the marker's .construct function to produce an object with those
    attrs.

    By default, the .construct uses the empty_instance cantrip, which creates
    blank copies of objects and then sets attributes on them in order to bypass
    any side-effects of the class's __init__ function:

    >>> from . import Int
    >>> class Pair(object):
    ...     def __init__(self, x, y):
    ...         print("Creating a pair!")
    ...         self.x, self.y = x, y
    ...     def __repr__(self):
    ...         return 'Pair({!r}, {!r})'.format(self.x, self.y)
    >>> int_pair = ObjectMarker(Pair).of(
    ...     x = Int(),
    ...     y = Int(),
    ... )
    >>> p1 = Pair(2,4)
    Creating a pair!
    >>> p2 = undictify(int_pair, dict(x=2, y=4))
    >>> (p2.x, p2.y) == (p1.x, p1.y)
    True
    >>> undictify(int_pair, (2,4))
    Traceback (most recent call last):
        ...
    Invalid: type_error - Expected a dict, got <type 'tuple'> instead

    You can also give an ObjectMarker a custom construct, which must accept
    keyword arguments of all the typegraph keys.

    >>> int_pair2 = ObjectMarker(Pair, constructor=Pair).of(
    ...     x = Int(),
    ...     y = Int()
    ... )
    >>> p3 = undictify(int_pair2, dict(x=2, y=4))
    Creating a pair!
    >>> (p3.x, p3.y) == (p1.x, p1.y)
    True

    Validate and traverse work as you'd expect:

    >>> validate(int_pair, p3)
    >>> invalid_p = Pair("hi", "ho")
    Creating a pair!
    >>> validate(int_pair, invalid_p)
    Traceback (most recent call last):
        ...
    Invalid: x: [type_error], y: [type_error]
    >>> validate(int_pair, (4,5))
    Traceback (most recent call last):
        ...
    Invalid: type_error - Expected Pair, got tuple

    >>> print(vg.ascii_tree(traverse(int_pair, p3), sort=True))
    root: Pair(2, 4)
      +--x: 2
      +--y: 4


    '''
    target_cls = None

    def construct(self, object_kwargs, **kwargs):
        if self.constructor:
            return self.constructor(**object_kwargs)
        return create_instance(self.target_cls, object_kwargs)

    def __init__(self, target_cls=None, constructor=None):
        if target_cls is not None:
            self.target_cls = target_cls
        self.constructor = constructor

    def of(self, **kwargs):
        children = {key:to_typegraph(val) for key, val in kwargs.items()}
        return vg.PlainGraphNode(self, **children)

def _as_dict(dispgraph, value, raise_invalid=False):
    result = {}
    for attr in dispgraph.key_iter():
        if raise_invalid and not hasattr(value, attr):
            raise Invalid("missing_attr", "Missing attribute {}".format(attr))
        result[attr] = getattr(value, attr)
    return result

@traverse.when(ObjectMarker)
def traverse_obj(dispgraph, value, zipgraph=None, **kwargs):
    d = _as_dict(dispgraph, value)
    g = dispgraph.super(ObjectMarker)(d, zipgraph, **kwargs)
    g.value = value if zipgraph is None else (value, zipgraph.value)
    return g

@validate.when(ObjectMarker)
def validate_obj(dispgraph, value, **kwargs):
    marker = dispgraph.marker
    if not isinstance(value, marker.target_cls):
        name = type(value).__name__
        expected = marker.target_cls.__name__
        raise Invalid("type_error", "Expected {}, got {}".format(expected, name))
    d = _as_dict(dispgraph, value, True)
    return dispgraph.super(ObjectMarker)(d, **kwargs)

@dictify.when(ObjectMarker)
def dictify_obj(dispgraph, value, **kwargs):
    d = _as_dict(dispgraph, value)
    return dispgraph.super(ObjectMarker)(d, **kwargs)

@undictify.when(ObjectMarker)
def undictify_obj(dispgraph, value, **kwargs):
    marker = dispgraph.marker
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    with error_agg.checking():
        result = dispgraph.super(ObjectMarker)(value, **kwargs)
    error_agg.raise_if_any()
    extra_keys = set(value.keys()) - set(dispgraph.key_iter())
    if extra_keys:
        error_agg.own_error(Invalid('unexpected_fields', keys=extra_keys))
    error_agg.raise_if_any()
    return marker.construct(result, **kwargs)




import collections

class ImmutableDict(collections.Mapping):
    '''An immutable dictionary.

    For the most part, it behaves like a normal dictionary, except that you
    can't set items:

    >>> repr(ImmutableDict(a=5))
    "ImmutableDict({'a': 5})"
    >>> d = ImmutableDict(dict(a=1, b=2), x=12, y=14)
    >>> assert d == {'a':1, 'b':2, 'x':12, 'y':14}
    >>> assert len(d) == 4
    >>> d['foo'] = 12
    Traceback (most recent call last):
        ...
    TypeError: 'ImmutableDict' object does not support item assignment

    It also has the .overlay method, which behaves like .update() on a normal
    dict, but returns a new ImmutableDict:

    >>> d2 = d.overlay(dict(a=2, foo='bar'))
    >>> d2 == dict(a=2, b=2, x=12, y=14, foo='bar')
    True

    Overlay also provides a strip_none argument, which defaults to False; if it
    is explicitly set to True, then keys in the new dict that are set to None
    will be dropped (instead of replaced by None):

    >>> not_stripped = d.overlay(dict(a=None, b=None))
    >>> not_stripped == dict(a=None, b=None, x=12, y=14)
    True
    >>> stripped = d.overlay(dict(a=None, b=None), strip_none=True)
    >>> stripped == dict(x=12, y=14)
    True

    The .select method creates a subset of this dict:

    >>> subd = d.select(['a', 'x'])
    >>> subd == dict(a=1, x=12)
    True

    '''
    def __init__(self, *args, **kwargs):
        self.__dict = dict(*args, **kwargs)

    def __getitem__(self, key):
        return self.__dict[key]

    def __contains__(self, key):
        return key in self.__dict

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __repr__(self):
        return "ImmutableDict({!r})".format(self.__dict)

    def overlay(self, other, strip_none=False):
        return overlay(self, other, strip_none)

    def select(self, keys):
        return select(self, keys)

    def omit(self, keys):
        return omit(self, keys)

def overlay(old, new, strip_none=False):
    '''
    Create a copy of one dict with values from a new one.

    >>> d = ImmutableDict(a=1, b=2)
    >>> d2 = overlay(d, dict(a=2, foo='bar'))
    >>> assert d2 == {'a':2, 'b':2, 'foo':'bar'}

    If the strip_none argument is True, keys in the new dict that are None will
    be dropped instead of replaced by None:

    >>> not_stripped = overlay(d, dict(a=None))
    >>> not_stripped == dict(a=None, b=2)
    True
    >>> stripped = overlay(d, dict(a=None), strip_none=True)
    >>> stripped == dict(b=2)
    True
    '''
    d = dict(old)
    d.update(new)
    if strip_none:
        for key, value in new.items():
            if value is None:
                del d[key]
    return ImmutableDict(d)

def select(old, keys):
    '''
    Create a subdict of an existing dict.

    >>> d = ImmutableDict(a=1, b=2, c=3, d="four")
    >>> d2 = select(d, {'a', 'd'})
    >>> assert d2 == dict(a=1, d="four")
    '''
    return ImmutableDict({k:old[k] for k in keys})

def omit(old, keys):
    '''
    Create a subdict of an existing dict by omitting keys

    >>> d = ImmutableDict(a=1, b=2, c=3, d="four")
    >>> d2 = omit(d, {'a', 'd'})
    >>> assert d2 == dict(b=2, c=3)
    '''
    return ImmutableDict({k:old[k] for k in set(old)-set(keys)})



import sys
if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str

from uuid import uuid4

from travesty import SchemaObj, String, Invalid, InvalidAggregator, ObjectMarker
from travesty import dictify, undictify, validate, traverse
from travesty.cantrips import empty_instance

def make_uid():
    '''Generate a (probably) unique id.'''
    return unicode(uuid4())

class UnloadedDocumentException(Exception):
    '''Raised on attr access of an unloaded document.'''
    def __init__(self, document, *a, **kw):
        self.document = document
        super(UnloadedDocumentException, self).__init__(document, *a, **kw)

class DoubleLoadException(Exception):
    '''Raised if a loaded document is loaded again.'''
    pass

class Document(SchemaObj):
    '''Document protocol for non-tree object graphs.

    Every document has a uid, which is a string. Each document is assumed to
    have a tree-like typegraph, but documents can contain references to other
    documents (or even to themselves), and this reference structure not be

    Every document has a uid; subclasses determine what other data they have.

    A Document object can be either loaded or unloaded; the doc.loaded flag
    indicates which a particular one is. A loaded document has data, and behaves
    normally; an unloaded document is a placeholder for a document that is
    stored in some external data store (such as a database) but has not had its
    data loaded.

    An unloaded Document has loaded=False; its uid is set but all the other
    attributes defined by field_types are missing, and trying to access them
    will raise an UnloadedDocumentException.

    In general you will never need to create unloaded documents directly, but
    the undictifier (see below) sometimes will.

    If no uid is provided during initialization, then the Document is given a
    generated uid and has new=True.
    '''
    field_types = dict(
        uid = String(),
    )
    loaded = False

    def __init__(self, uid=None, **attrs):
        self.loaded = True
        super(Document, self).__init__()
        # If uid is not specified, generate one randomly
        if uid is None:
            uid = make_uid()
        self.uid = uid
        self._load(**attrs)

    @classmethod
    def _create_unloaded(cls, uid):
        '''Create an unloaded instance of this class.'''
        return empty_instance.create_instance(cls, dict(
            loaded = False,
            uid = uid,
        ))

    def load(self, **attrs):
        if self.loaded:
            raise DoubleLoadException(type(self), self.uid)
        self._load(**attrs)
        return self

    def _load(self, **attrs):
        '''Load an unloaded instance of this class'''
        self.loaded = True
        extras = set(attrs.keys()) - set(self.field_types.keys())
        if extras:
            keylist = ', '.join('"{}"'.format(k) for k in extras)
            msg = "{0}() got an unexpected keyword arguments {1}"
            msg = msg.format(type(self).__name__, keylist)
            raise TypeError(msg)
        for key in self.field_types.keys():
            if key == 'uid': continue
            setattr(self, key, attrs.get(key, None))
        return self

    def __getattr__(self, attr):
        '''Raise UnloadedDocumentException on access to unloaded attributes'''
        if not self.loaded and attr != 'uid' and attr in self.field_types:
            raise UnloadedDocumentException(self)
        return object.__getattribute__(self, attr)

    def _loaded_str(self):
        '''__str__ for loaded instances.

        Subclassess can override this to specify __str__ behavior for loaded
        instances while preserving the default for unloaded instances.
        '''
        return "<{}: {}>".format(type(self).__name__, self.uid)

    def __str__(self):
        if self.loaded:
            return self._loaded_str()
        return "<Unloaded {}: {}>".format(type(self).__name__, self.uid)

    def _loaded_repr(self):
        '''__repr__ for loaded instances.

        Subclassess can override this to specify __repr__ behavior for loaded
        instances while preserving the default for unloaded instances.
        '''
        return "<{}: {}>".format(type(self).__name__, self.uid)

    def __repr__(self):
        if self.loaded:
            return self._loaded_repr()
        return str(self)


@dictify.when(Document.marker_cls)
def df_document(dispgraph, doc, doc_storage=None, no_doc_kids=False, **kwargs):
    '''Dictify a document.

    The behavior is different depending on whether doc_storage kwarg is None.

    If it is, then this behaves just like dictifying a SchemaObj - the document
    is serialized to a dictionary, which is returned.

    If doc_storage is not None, then it should be a dictionary mapping uids to
    serialized Documents. In this case, this function will always return
    {'uid':doc.uid}, but if doc.uid isn't already in doc_storage then the
    serialized dictionary of this object will be added to it as a side effect.

    Thus, if you have a complex structure of Documents, dictifying the root with
    no doc_storage may give you multiple serialized copies of the same object,
    while serializing with doc_storage={} will populate doc_storage with
    serialized versions of each object, leaving only {'uid':<uid>} where
    sub-documents are present.

    '''
    # unloaded doc -> {"uid":doc.uid}
    if not doc.loaded or no_doc_kids == "internal":
        return dict(uid=doc.uid)
    # doc already stored -> {"uid": doc.uid}
    if doc_storage is not None and doc.uid in doc_storage:
        return dict(uid=doc.uid)
    if doc_storage is not None: # place a marker in here in case of recursion
        doc_storage[doc.uid] = dict(uid=doc.uid)
    superdisp = dispgraph.super(Document.marker_cls)
    if no_doc_kids:
        no_doc_kids = "internal"
    result = superdisp(doc, doc_storage=doc_storage, no_doc_kids=no_doc_kids, **kwargs)
    if doc_storage is not None:
        # When using storage, always return just {'uid':doc.uid}
        doc_storage[doc.uid] = result
        return dict(uid=doc.uid)
    # without storage, return full result
    return result


@undictify.when(Document.marker_cls)
def udf_document(dispgraph, value, in_docset=None, **kwargs):
    if not isinstance(value, dict):
        raise Invalid('type_error', "Expected dict, got {}".format(type(value)))
    if 'uid' not in value:
        raise Invalid('missing_key:uid', "Document has no uid.")
    uid = value['uid']
    doctype = dispgraph.marker.target_cls
    if in_docset is not None:
        doc = in_docset.get_or_create(doctype, uid)
    else:
        doc = doctype._create_unloaded(uid)
    # If the input has no keys besides 'uid', and the doctype expects more, then
    # this is an unloaded document and we should just return it
    if (len(value) == 1) and len(doctype.field_types) > 1:
        return doc
    # Otherwise, we need to populate it.
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    with error_agg.checking():
        # Need to jump over ObjectMarker and go straight to Schema so we get a
        # dictionary back
        # TODO could probably refactor ObjectMarker and/or Schema to avoid this
        result = dispgraph.super(ObjectMarker)(value, in_docset=in_docset, **kwargs)
    extra_keys = set(value.keys()) - set(dispgraph.key_iter())
    if extra_keys:
        error_agg.own_error(Invalid('unexpected_fields', keys=extra_keys))
    error_agg.raise_if_any()
    doc.load(**result)
    return doc

@validate.when(Document.marker_cls)
def vd_document(dispgraph, doc, no_doc_kids=False, **kwargs):
    doctype = dispgraph.marker.target_cls
    if not isinstance(doc, doctype):
        raise Invalid("type_error", "Expected {}, got {}".format(doctype, type(doc)))
    # This actually works, amusingly enough
    traverse_document(dispgraph, doc, no_doc_kids=no_doc_kids, **kwargs)

@traverse.when(Document.marker_cls)
def traverse_document(dispgraph, doc, zipgraph=None, no_doc_kids=False, **kwargs):
    superdisp = dispgraph.super(Document.marker_cls)
    if not doc.loaded or no_doc_kids == "internal":
        # Unloaded doc: only traverse uid
        return superdisp.restrict(['uid'])(doc, zipgraph=zipgraph, no_doc_kids=no_doc_kids, **kwargs)
    if no_doc_kids:
        no_doc_kids = "internal"
    # Treat loaded docs like any other SchemaObj
    return superdisp(doc, zipgraph=zipgraph, no_doc_kids=no_doc_kids, **kwargs)



'''
A collection of utilities that aren't large enough to have their own modules.

Each file in this directory is a useful standalone tool, that may or may not
someday be moved into its own separate package.
'''


class _EmptyOldClass:
    '''Placeholder old style class assembling blank instances.'''
    pass


class _EmptyNewClass(object):
    '''Placeholder new style class assembling blank instances.'''
    pass


def empty_instance(cls):
    '''Create an instance of cls without invoking cls.__init__.

    Supports both old- and new-style classes:

    >>> class Foo:
    ...     def __init__(self, x):
    ...         self.x = x
    >>> class Bar(object):
    ...     def __init__(self, x):
    ...         self.x = x
    >>> foo = empty_instance(Foo)
    >>> bar = empty_instance(Bar)
    >>> hasattr(foo, 'x')
    False
    >>> hasattr(bar, 'x')
    False

    '''
    if issubclass(cls, object):
        instance = _EmptyNewClass()
    else: # pragma: no cover
        instance = _EmptyOldClass()
    instance.__class__ = cls
    return instance

def create_instance(cls, kwargs):
    '''Create and populate an instance of cls.

    >>> class Bar(object): pass
    >>> b = create_instance(Bar, dict(x=12))
    >>> b.x
    12
    '''
    instance = empty_instance(cls)
    for key, val in kwargs.items():
        setattr(instance, key, val)
    return instance

def ctor(cls):
    '''Return a constructor function for cls.

    >>> class Bar(object): pass
    >>> make_bar = ctor(Bar)
    >>> b = make_bar(x=12, y=14)
    >>> b.x
    12
    >>> b.y
    14

    '''
    return lambda **kwargs: create_instance(cls, kwargs)


from __future__ import absolute_import

import vertigo as vg

from .cantrips.immutable_dict import ImmutableDict
from .cantrips.subclass import Subclassable

from .base import Traversable, to_typegraph
from .object_marker import ObjectMarker

def _marker_for(cls):
    '''Get a marker for a SchemaObj subclass.

    The returned marker will be a subclass of ObjectMarker targeting the input
    class.
    '''
    bases = []
    for base in cls.__bases__:
        if issubclass(base, SchemaObj):
            bases.append(base.marker_cls)
    if not bases:
        bases = [ObjectMarker]
    return type(cls.__name__+"Marker", tuple(bases), dict(target_cls=cls))

class SchemaObj(Traversable, Subclassable):
    '''Type for making python classes with automatically-inferred typegraphs.

    The class attribute field_types should be a dictionary mapping attribute
    names to typegraphs (or valid targets for to_typegraph). Some metaclass
    magic will automatically generate the class attributes .marker_cls and
    .typegraph, where .marker_cls is an ObjectMarker subclass pointing at this
    type, and .typegraph is a typegraph whose root contains an instance of
    cls.marker_cls and whose children are determined by cls.field_types.

    To support recursive types, you can specify a single-argument callable for
    field_types, which takes the class and returns a valid field_types dict. If
    you use this method, the class will also get a classmethod called
    finalize_typegraph which MUST be called before you do anything significant
    with the class. Before this function is called, cls.typegraph will have no
    children; when you call finalize_typegraph, the children will be populated
    and the finalize_typegraph method will be removed.
    '''
    field_types = ImmutableDict()

    def __init__(self, **kwargs):
        for key in self.field_types:
            if key not in kwargs:
                kwargs[key] = None
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def __subclass__(cls, field_types=None, **kwargs):
        super(SchemaObj, cls).__subclass__(**kwargs)
        cls.marker_cls = _marker_for(cls)
        # It is CRUCIAL that we set cls.typegraph now instead of in
        # finalize_typegraph in case something needs the graph before then.
        cls.typegraph = vg.PlainGraphNode(cls.marker_cls())
        if callable(field_types):
            @classmethod
            def finalize_typegraph(cls):
                cls.__construct_typegraph(field_types(cls))
                del cls._finalize_typegraph
            cls._finalize_typegraph = finalize_typegraph
        else:
            cls.__construct_typegraph(field_types)

    @classmethod
    def __construct_typegraph(cls, field_types):
        if field_types is None:
            field_types = {}
        cls._field_types = field_types
        cls.field_types = cls.field_types.overlay(field_types, strip_none=True)
        for key, value in cls.field_types.items():
            cls.typegraph[key] = to_typegraph(value)


# Force SchemaObj to set itself up
SchemaObj.__subclass__()




from travesty import undictify

from .document import DoubleLoadException


class DocSet(object):
    '''Tool for loading linked documents.

    doc_loader.get_or_create(type, uid) returns a document of the given type
    with the given uid.

    doc_loader.load(type, uid, data) works like get_or_create, but loads the
    document from the given data.

    Because the doc_loader keeps only one object for each uid, calling
    get_or_create() twice will return the same object, and more importantly, if
    you call get_or_create() and then later call load() with the same type and
    uid, both will return the same object, and so after calling load() the
    object returned by get_or_create() will now be loaded:

    >>> from travesty import String
    >>> from . import Document
    >>> class Note(Document):
    ...     field_types = dict(content=String())
    >>> docset = DocSet()
    >>> note = docset.get_or_create(Note, 'some_uid')
    >>> note.content
    Traceback (most recent call last):
        ...
    UnloadedDocumentException: <Unloaded Note: some_uid>
    >>> serialized_note = {'uid':'some_uid', 'content':'Lorem ipsum etc.'}
    >>> note2 = docset.load(Note, serialized_note)
    >>> note2.content
    u'Lorem ipsum etc.'
    >>> note.content
    u'Lorem ipsum etc.'
    >>> note is note2
    True

    The undictify implementation for Document can take a DocSets as a
    keyword arg. If the keyword arg 'in_docset' is a DocSet d, then Documents
    will be loaded via d.get_or_create (instead of created unloaded), and will
    be stored in d as they are loaded.

    The helper method docset.load(type, data) is shorthand for
    travesty.undictify(type, data, in_docset=docset)

    '''
    def __init__(self, items=()):
        #: (schema_cls, uid) -> document
        self.document_map = {}
        for item in items:
            self.add(item)

    def add(self, doc):
        key = (type(doc), doc.uid)
        if key in self.document_map:
            raise ValueError("Duplicate uid {1} for type {0}".format(*key))
        self.document_map[key] = doc

    def get(self, type, uid):
        key = (type, uid)
        return self.document_map.get(key, None)

    def create(self, type, uid):
        key = (type, uid)
        if key in self.document_map:
            raise ValueError("Duplicate uid {1} for type {0}".format(*key))
        doc = type._create_unloaded(uid)
        self.document_map[key] = doc
        return doc

    def get_or_create(self, type, uid):
        '''Return a document with the given type and uid

        If this is called multiple times with the same uid, the same document
        will be returned.
        '''
        key = (type, uid)
        doc = self.document_map.setdefault(key, type._create_unloaded(uid))
        assert isinstance(doc, type)
        return doc

    def load(self, type, data, allow_double_load=False, kwargs=None):
        '''Return a loaded document with the given type and uid.

        This returns the same object as a call to get_or_create(type, uid); if
        you call get_or_create and then load, load will return the same object
        but will also populate it with data.

        This is precisely equivalent to

        undictify(type, data, in_docset=self, **kwargs)

        Note that you cannot choose to use a different typegraph or dispatcher
        here - if you need to, just call it directly.
        '''
        if isinstance(data, dict) and 'uid' in data:
            old = self.get(type, data['uid'])
            if old and old.loaded:
                if allow_double_load:
                    return old
                raise DoubleLoadException(type, data['uid'])
        kwargs = kwargs.copy() if kwargs else {}
        kwargs['in_docset'] = self
        return undictify(type, data, **kwargs)

    def __getitem__(self, key):
        return self.document_map[key]

    def __contains__(self, key):
        return key in self.document_map



import vertigo as vg

from .base import Marker, Traversable, to_typegraph
from .base import traverse, validate, dictify, undictify
from .invalid import Invalid

class UnknownType(Exception):
    '''Raised when a Polymorph encounters an unknown type.'''
    pass

class Polymorph(Marker):
    '''The following docstring is ENTIRELY SUSPECT. TODO: Fix it.

    A dictifier that selects between several dictifiables based on classes.

    The mapping argument is a dict whose keys are strings and whose values are
    either (class_list, Dictifier) pairs or Dictifiable classes.

    Polymorph can Dictify any object that is an instance of one of the
    Dictifiable classes or any class in one of the class_list entries; it will
    dictify it as a tuple of (name, value), where name identifies which
    Dictifier was used and value is the dictified object; on undictifiance, it
    undictifies the value using the Dictifier for that name.

    For example:

    >>> from . import Number, Complex, SchemaObj
    >>> class Point(SchemaObj):
    ...     field_types = dict(x=Number(), y=Complex())
    ...     def __repr__(self): return "Point({}, {})".format(self.x, self.y)
    ...
    >>> num_or_point = Polymorph.mkgraph({
    ...     'point':Point,
    ...     'num':((int, float), Number()),
    ... })
    >>> dictify(num_or_point, 12)
    ('num', 12)
    >>> dictify(num_or_point, 3.5)
    ('num', 3.5)
    >>> dictify(num_or_point, Point(x=1, y=3)) == ('point', {'x':1, 'y':3})
    True
    >>> undictify(num_or_point, ('num', 42))
    42
    >>> p = undictify(num_or_point, ('point', {'x':-1, 'y':2+2j}))
    >>> p.x
    -1
    >>> p.y
    (2+2j)

    The input to undictify MUST be a list or tuple of length 2:

    >>> undictify(num_or_point, 42)
    Traceback (most recent call last):
        ...
    Invalid: type_error
    >>> undictify(num_or_point, ('num', 42, 'something extra for no reason'))
    Traceback (most recent call last):
        ...
    Invalid: bad_list

    It will validate only the types in its list:

    >>> validate(num_or_point, 3)
    >>> validate(num_or_point, 3.5)
    >>> validate(num_or_point, p)
    >>> validate(num_or_point, 'hi')
    Traceback (most recent call last):
    ...
    Invalid: type_error - Unrecognized type: <type 'str'>

    Traversal with a zipgraph expects the graph to have keys matching the main
    graph structure, as always, but note that the graph value at the actual
    Polymorph will be ignored

    >>> g = vg.from_dict(dict(_self="Ignored", num="A number", point=dict(
    ... _self="A Point", x="The X", y="The Y")))
    >>> print(vg.ascii_tree(traverse(num_or_point, 3, zipgraph=g)))
    root: (3, 'A number')
    >>> print(vg.ascii_tree(traverse(num_or_point, p, zipgraph=g), sort=True))
    root: (Point(-1, (2+2j)), 'A Point')
      +--x: (-1, 'The X')
      +--y: ((2+2j), 'The Y')

    Note that the initialization above is short for the longer form:
    >>> num_or_point = Polymorph({'point':Point, 'num':(int, float)}).of(
    ...     point = Point, num=Number()
    ... )


    '''
    def __init__(self, mapping):
        super(Polymorph, self).__init__()
        self.cls_to_name = {}
        for name, classes in mapping.items():
            if not isinstance(classes, (list, tuple)):
                classes = (classes,)
            for cls in classes:
                self.cls_to_name[cls] = name

    def name_for_type(self, typ):
        '''Given a type, get the corresponding name.'''
        for basetype in typ.__mro__:
            if basetype in self.cls_to_name:
                return self.cls_to_name[basetype]
        raise UnknownType(typ.__name__)

    def name_for_val(self, value):
        '''Given an object, get the corresponding name.'''
        return self.name_for_type(type(value))

    def of(self, **kwargs):
        children = {key:to_typegraph(val) for key, val in kwargs.items()}
        return vg.PlainGraphNode(self, **children)

    @classmethod
    def mkgraph(cls, mapping):
        # Mapping should be name:(type_or_types, typegraph) or name:Traversable
        lookup = {}
        children = {}
        for name, data in mapping.items():
            if isinstance(data, type) and issubclass(data, Traversable):
                lookup[name] = (data,)
                children[name] = data.typegraph
            else:
                lookup[name], children[name] = data
        return cls(lookup).of(**children)


@undictify.when(Polymorph)
def undictify_pmorph(dispgraph, value, **kwargs):
    '''Convert a dictified value to an object and return it.'''
    if not isinstance(value, (list, tuple)):
        raise Invalid('type_error')
    if len(value) != 2:
        raise Invalid('bad_list')
    name, value = value
    return dispgraph[name](value, **kwargs)

@dictify.when(Polymorph)
def dictify_pmorph(dispgraph, value, **kwargs):
    name = dispgraph.marker.name_for_val(value)
    return (name, dispgraph[name](value, **kwargs))

@validate.when(Polymorph)
def validate_pmorph(dispgraph, value, **kwargs):
    try:
        name = dispgraph.marker.name_for_val(value)
    except UnknownType:
        raise Invalid("type_error", "Unrecognized type: {}".format(type(value)))
    dispgraph[name](value, **kwargs)

@traverse.when(Polymorph)
def traverse_pmorph(dispgraph, value, zipgraph=None, **kwargs):
    name = dispgraph.marker.name_for_val(value)
    zg = zipgraph[name] if zipgraph else None
    return dispgraph[name](value, zg, **kwargs)



import vertigo as vg

from .invalid import Invalid, InvalidAggregator
from .base import Marker, traverse, validate, dictify, undictify, to_typegraph

class Schema(Marker):
    '''Abstract Marker for schema-structured objects.'''

    def of(self, **kwargs):
        children = {key:to_typegraph(val) for key, val in kwargs.items()}
        return vg.PlainGraphNode(self, **children)

_foo = traverse

@traverse.when(Schema)
def traverse_schema(dispgraph, value, zipgraph=None, **kwargs):
    edges = []
    for key, subgraph in dispgraph.edge_iter():
        subzip = zipgraph[key] if zipgraph else None
        edges.append((key, subgraph(value[key], subzip, **kwargs)))
    v = value
    if zipgraph:
        v = (v, zipgraph.value)
    return vg.PlainGraphNode(v, edges)

@validate.when(Schema)
def validate_schema(dispgraph, value, **kwargs):
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    for key, subgraph in dispgraph.edge_iter():
        with error_agg.checking_sub(key):
            if key not in value:
                raise Invalid("missing_key", "Missing key {}".format(key))
            subgraph(value[key], **kwargs)
    error_agg.raise_if_any()

@dictify.when(Schema)
def dictify_schema(dispgraph, value, **kwargs):
    result = {}
    for key, subgraph in dispgraph.edge_iter():
        result[key] = subgraph(value[key], **kwargs)
    return result

@undictify.when(Schema)
def undictify_schema(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid('type_error', 'Expected a dict, got {} instead'.format(type(value)))
    # If fail_early is True, then gather all errors from this and its
    # children. Otherwise, just raise the first error we encounter.
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    result = {}
    for key, subgraph in dispgraph.edge_iter():
        with error_agg.checking_sub(key):
            val = value.get(key, None)
            result[key] = subgraph(val, **kwargs)
    error_agg.raise_if_any()
    return result



'''Tools for inferring the best marker for an object

>>> infer_marker(True)
<Boolean>
>>> infer_marker(12)
<Int>
>>> infer_marker("hello")
<String>
>>> infer_marker([1,2,3])
Traceback (most recent call last):
    ...
NotImplementedError: <type 'list'>
'''
import datetime
import sys

from .cantrips.dispatcher import Dispatcher
from . import Boolean, Complex, Date, DateTime, Int, Number, String, Time

if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str


basic_mapping = {
    bool: Boolean(),
    complex: Complex(),
    datetime.date: Date(),
    datetime.time: Time(),
    datetime.datetime: DateTime(),
    int: Int(),
    float: Number(),
    basestring: String(),
}

infer_marker = Dispatcher()

@infer_marker.default()
def infer_basic(d, value):
    for typ in type(value).__mro__:
        if typ in basic_mapping:
            return basic_mapping[typ]
    msg = "Cannot infer marker for type {}."
    raise NotImplementedError(msg.format(type(value)))


# TODO infer_typegraph for more complex structures?
# We'll see if we have a use for that.


"""
Markers and core functions for dates, times, and datetimes.

All three types assume you're using the date, time, and datetime types from the
datetime module.

This module exposes three Markers: DateTime, Date, and Time

validate accepts only datetime, date, or time objects, respectively:

>>> now = datetime.datetime.now()
>>> validate(DateTime(), now)
>>> validate(DateTime(), "")
Traceback (most recent call last):
...
Invalid: type_error

>>> validate(Date(), now.date())
>>> validate(Date(), now.time())
Traceback (most recent call last):
...
Invalid: type_error

>>> validate(Time(), now.time())
>>> validate(Time(), now)
Traceback (most recent call last):
...
Invalid: type_error


Their dictify implementations use ISO 8601 format:

>>> d = datetime.datetime(1776, 12, 25, 14, 21, 3, 100500)
>>> dictify(DateTime(), d)
u'1776-12-25T14:21:03.100500'
>>> dictify(Date(), d.date())
u'1776-12-25'
>>> dictify(Time(), d.time())
u'14:21:03.100500'
>>> dictify(DateTime(), d.replace(microsecond=0))
u'1776-12-25T14:21:03'


>>> undictify(DateTime(), u'1776-12-25T14:21:03.100500')
datetime.datetime(1776, 12, 25, 14, 21, 3, 100500)
>>> undictify(Date(), u'1776-12-25')
datetime.date(1776, 12, 25)
>>> undictify(Time(), u'14:21:03.100500')
datetime.time(14, 21, 3, 100500)
>>> undictify(DateTime(), u'1776-12-25T14:21:03')
datetime.datetime(1776, 12, 25, 14, 21, 3)

Unparseable strings get "bad_format" errors; non-strings get "type_error":

>>> undictify(DateTime(), "20003-1-1")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), "1776-13-25")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), "1776-12-32")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), "1776-12-25T26:25:03.1")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), "1776-12-25T14:71:03.1")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), "1776-12-25T14:25:93.1")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(Date(), "1776-hello-12")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(Time(), "Not even remotely a datetime.")
Traceback (most recent call last):
...
Invalid: bad_format
>>> undictify(DateTime(), ["this", "isn't", "even", "a", "string"])
Traceback (most recent call last):
...
Invalid: type_error



Like all good undictify implementations, it is idempotent:

>>> dt = undictify(DateTime(), "1776-12-25T14:21:03.100500")
>>> undictify(DateTime(), dt) == dt
True
>>> undictify(Date(), dt.date()) == dt.date()
True
>>> undictify(Time(), dt.time()) == dt.time()
True
"""



import datetime

import sys
if sys.version >= '3': # pragma: no cover
    unicode = str
    basestring = str

from .base import Leaf, validate, dictify, undictify
from .invalid import Invalid

class _DateTimeMarker(Leaf):
    _dt_type = None
    _format = None

@validate.when(_DateTimeMarker)
def validate_dt(dispgraph, value, **kwargs):
    if not isinstance(value, dispgraph.marker._dt_type):
        raise Invalid("type_error")

@dictify.when(_DateTimeMarker)
def dictify_dt(dispgraph, value, **kwargs):
    return unicode(value.isoformat())

@undictify.when(_DateTimeMarker)
def undictify_dt(dispgraph, value, **kwargs):
    marker = dispgraph.marker
    return _parse(value, marker._dt_type)

class DateTime(_DateTimeMarker):
    _dt_type = datetime.datetime

class Date(_DateTimeMarker):
    _dt_type = datetime.date

class Time(_DateTimeMarker):
    _dt_type = datetime.time

class TimeDelta(Leaf):
    pass

@validate.when(TimeDelta)
def validate_td(dispgraph, value, **kwargs):
    if not isinstance(value, datetime.timedelta):
        raise Invalid("type_error")

@dictify.when(TimeDelta)
def dictify_dt(dispgraph, value, **kwargs):
    return (value.minutes, value.seconds, value.microseconds)

@undictify.when(TimeDelta)
def undictify_dt(dispgraph, value, **kwargs):
    if not isinstance(value, (list, tuple)):
        raise Invalid("type_error", "Expected list or tuple, got {}".format(type(value)))
    if len(value != 3):
        raise Invalid("value_error", "Expected 3 items, not {}".format(len(value)))
    try:
        return dt.timedelta(*value)
    except TypeError as e:
        raise Invalid("type_error", str(e))

def _parse(value, cls=datetime.datetime):
    '''
    >>> _parse("2001-01-01T5:10:15")
    datetime.datetime(2001, 1, 1, 5, 10, 15)
    >>> _parse("2001-01-01", datetime.date)
    datetime.date(2001, 1, 1)
    >>> _parse("5:10:15", datetime.time)
    datetime.time(5, 10, 15)
    >>> _parse("5:10:15.42", datetime.time)
    datetime.time(5, 10, 15, 420000)
    >>> _parse("5:10:15.42", datetime.timedelta)
    Traceback (most recent call last):
        ...
    ValueError: Unrecognized date type: <type 'datetime.timedelta'>
    '''
    if cls is datetime.datetime:
        fmt = "%Y-%m-%dT%H:%M:%S"
    elif cls is datetime.date:
        fmt = "%Y-%m-%d"
    elif cls is datetime.time:
        fmt = "%H:%M:%S"
    else:
        raise ValueError("Unrecognized date type: {}".format(cls))
    if isinstance(value, cls):
        return value
    if not isinstance(value, basestring):
        raise Invalid("type_error")
    if '.' in value and cls in [datetime.datetime, datetime.time]:
        fmt = fmt + ".%f"
    try:
        val = datetime.datetime.strptime(value, fmt)
    except ValueError as e:
        raise Invalid("bad_format", str(e))
    if cls is datetime.datetime:
        return val
    elif cls is datetime.date:
        return val.date()
    else:
        return val.time()



'''subclass.py: mixin for programmatically subclassable classes.'''
from abc import ABCMeta

class SubclassMixin(object):
    '''SubclassMixin adds a subclass(**kwargs) method to a class.

    cls.subclass(**kwargs) is equivalent to creating a new subclass with the
    contents of kwargs written out within it.

    For example:

    >>> class Point(SubclassMixin):
    ...   round = False
    ...   def __init__(self, x, y):
    ...     if self.round:
    ...        x, y = int(x), int(y)
    ...     self.x, self.y = x, y
    ...
    >>> p = Point(.5, 3.25)
    >>> p.x, p.y
    (0.5, 3.25)

    The following two subclasses are (nearly) equivalent

    >>> class IntPoint1(Point):
    ...   round = True
    ...
    >>> IntPoint2 = Point.subclass(round=True)

    They behave exactly the same:

    >>> p1 = IntPoint1(.5, 3.25)
    >>> p2 = IntPoint2(.5, 3.25)
    >>> p1.x, p1.y
    (0, 3)
    >>> p2.x, p2.y
    (0, 3)

    The only difference is their names, as IntPoint2's name was auto-generated:

    >>> IntPoint1.__name__
    'IntPoint1'
    >>> IntPoint2.__name__
    'Point_sub'

    This can be rectified via the special __class_name attribute:

    >>> IntPoint3 = Point.subclass(round=True, __class_name="IntPoint1")
    >>> IntPoint3.__name__
    'IntPoint1'

    '''
    @classmethod
    def subclass(cls, **kwargs):
        '''
        Programmatically generate subclasses of this class.

        The generated subclass will be called <name of this class>_sub unless
        the special parameter __class_name is passed in.
        '''
        if '__class_name' in kwargs:
            class_name = kwargs.pop('__class_name')
        else:
            class_name = cls.__name__ + "_sub"
        subcls = type(cls)(class_name, (cls,), kwargs)
        return subcls

    @classmethod
    def __subclass__(cls, **kwargs):
        for name, val in kwargs.items():
            setattr(cls, name, val)


def _im_func(fn): # pragma: no cover
    if hasattr(fn, 'im_func'):
        return fn.im_func
    if hasattr(fn, '__func__'):
        return fn.__func__
    return fn

class SubclassableMeta(ABCMeta):
    '''Metaclass for classes that do extra work on subclassing.

    Classes with this metaclass can define a __subclass__(cls, **kwargs) method
    that will be called on subclasses upon creation.

    For example, consider this class, whose __subclass__ renames the 'foo'
    attribute to 'bar':

    >>> class AutoBar(Subclassable):
    ...     def __subclass__(cls, foo=None, **kwargs):
    ...         kwargs.setdefault('bar', foo)
    ...         super(AutoBar, cls).__subclass__(**kwargs)

    Then we can create a subclass of AutoBar:

    >>> class Foo(AutoBar):
    ...     foo = 12

    And it will have a 'bar' attr instead of 'foo':

    >>> Foo.bar
    12
    >>> Foo.foo
    Traceback (most recent call last):
        ...
    AttributeError: type object 'Foo' has no attribute 'foo'


    Note that __subclass__ is NOT be called on the root class, so AutoBar itself
    has no 'bar' attr:

    >>> AutoBar.bar
    Traceback (most recent call last):
        ...
    AttributeError: type object 'AutoBar' has no attribute 'bar'
    '''
    def __new__(cls, name, supers, kwargs):
        t = ABCMeta.__new__(cls, name, supers, {})
        # Force __subclass__ to be a classmethod
        # if not isinstance(t.__subclass__, classmethod):
        #     t.__subclass__ = classmethod(_im_func(t.__subclass__))
        if '__subclass__' in kwargs:
            sc = kwargs['__subclass__']
            if not isinstance(sc, classmethod):
                kwargs['__subclass__'] = classmethod(sc)
        t.__subclass__(**kwargs)
        return t

Subclassable = SubclassableMeta('Subclassable', (SubclassMixin,), {})



import vertigo as vg

from . import Wrapper, traverse, dictify, undictify, validate

class Optional(Wrapper):
    '''Wrapper that indicates the value could be None.

    >>> from . import Int, List

    None fails validation, won't always undictify, etc.

    >>> validate(Int(), None)
    Traceback (most recent call last):
        ...
    Invalid: type_error
    >>> undictify(List().of(Int()), None)
    Traceback (most recent call last):
        ...
    Invalid: type_error

    Wrapping it in Optional fixes this:

    >>> validate(Optional.wrap(Int()), None)
    >>> undictify(Optional.wrap(Int()), None) is None
    True

    '''
    pass

@traverse.when(Optional)
def traverse_optional(dispgraph, value, zipgraph=None, **kwargs):
    if value is None:
        if zipgraph:
            return vg.PlainGraphNode((value, zipgraph.value))
        return vg.PlainGraphNode(value)
    opt = dispgraph.marker
    return dispgraph.for_marker(opt.marker)(value, zipgraph=zipgraph, **kwargs)


@dictify.when(Optional)
@undictify.when(Optional)
def dfy_undfy_optional(dispgraph, value, **kwargs):
    if value is None:
        return None
    opt = dispgraph.marker
    return dispgraph.for_marker(opt.marker)(value, **kwargs)


@validate.when(Optional)
def validate_optional(dispgraph, value, **kwargs):
    if value is None:
        return
    opt = dispgraph.marker
    dispgraph.for_marker(opt.marker)(value, **kwargs)



from .base import Marker, dictify, undictify

class Passthrough(Marker):
    '''A leaf value that can be dictified as is and needs no validation.

    >>> dictify(Passthrough(), 12)
    12
    >>> undictify(Passthrough(), {'x': [1, 2, 3]})
    {'x': [1, 2, 3]}
    '''

@dictify.when(Passthrough)
@undictify.when(Passthrough)
def passthrough(dispgraph, value, **kwargs):
    return value




import vertigo as vg

from .cantrips.dispatcher import Dispatcher, DispatchSuper, SuperMarker

#  =================
#  = DispatchGraph =
#  =================

# Adding special methods to GraphNodes is discouraged. I'm only doing it here
# because it makes the API MUCH simpler when writing external target functions,
# and in general we don't expect the recieving functions to manipulate these.
class DispatchGraph(object):
    '''Abstract base mixin for dispatch graphs.

    A DispatchGraph is just a wrapper around a graph-aware dispatcher and a
    marker value, with some helpful functions.

    Dispatch Graphs have several core attributes:

    .target - the object to dispatch on (any value or SuperMarker)
    .disp_target - the dispatcher to use (a Dispatcher or a DispatchSuper)
    .value - a tuple of (.target, .disp_target)
    .marker_graph() - the wrapped marker graph

    Subclasses must either provide .value OR both .target and .disp_target. The
    other will be inferred automatically.

    The .marker property exposes the marker value for this node, which is the
    same as .target unless .target is a SuperMarker, in which case .marker is
    the actual marker value that the SuperMarker is wrapping.

    Similarly, the .disp property is the same as .disp_target except when
    .disp_target is a DispatchSuper, when .disp is the root dispatcher instead.

    DispatchGraphs are callable - calling one will use its dispatcher and target
    to choose a function, and then invoke that function.
    '''
    __slots__ = ()

    def __call__(self, *args, **kwargs):
        fn = self._get_fn()
        if not fn:
            raise NotImplementedError(self.marker)
        return fn(self, *args, **kwargs)

    @property
    def target(self):
        return self.value[0]

    @property
    def disp_target(self):
        return self.value[1]

    @property
    def disp(self):
        if isinstance(self.disp_target, DispatchSuper):
            return self.disp_target.disp
        return self.disp_target

    @property
    def value(self):
        return (self.target, self.disp_target)

    @property
    def marker(self):
        if isinstance(self.target, SuperMarker):
            return self.target.val
        return self.target

    def marker_graph(self):
        raise NotImplementedError()

    def _get_fn(self):
        return self.disp_target.dispatch(self.target)

    def for_marker(self, marker):
        '''Overlay a new marker type on this graph.

        The returned graph is functionally identical to this one, but with a
        different marker type at this node.
        '''
        return DispatchOverlay(self, target=marker)

    def super(self, cls):
        '''Get a DispatchGraph for the supertype of this one's target.

        The returned graph is functionally identical to this one, but with a
        super object as its target. The type of the new target is obtained by
        walking up the current marker's __mro__. In particular, if the current
        target is a super instance super(SomeMarker, obj), this will sensibly
        choose the next supertype after SomeMarker in obj's __mro__.
        '''
        return self.for_marker(SuperMarker(cls, self.marker))

    def parent(self, disp):
        '''Get a DispatchGraph for the parent of this one's dispatcher.

        The returned graph is functionally identical to this one, but with a
        different dispatcher *at this node*. Subnodes will still have the
        original dispatcher.

        BUG: This stacks weirdly with for_marker: if you parent() up one level
        and call .for_marker(), you'd expect to get the child dispatcher in the
        result, but instead you'll still have the parent. The right way to do
        this probably involves something like super() - it behaves like the
        parent, but still is the original.

        TODO Fix this? Or maybe it's ok? Or maybe it'll never come up?
        '''
        return DispatchOverlay(self, disp_target=DispatchSuper(disp, self.disp))

    def inner(self, *args, **kwargs):
        '''Unwrap self.marker by one layer and overlap.

        This is shorthand for self.for_marker(self.marker.marker)(*args,
        **kwargs). If self.marker is a Wrapper, this will call the dispatcher on
        its inner marker. If self.marker is not a wrapper, this will fail and
        your program will crash and it will be all your fault, you monster.
        '''
        return self.for_marker(self.marker.marker)(*args, **kwargs)

    def restrict(self, edge_names):
        return DispatchRestriction(self, edge_names)

# Wraps a graph of Marker values plus a common dispatcher
class DynamicDispatchGraph(DispatchGraph, vg.wrappers.GraphWrapper):
    '''DispatchGraph that wraps a graph of Markers.'''
    __slots__ = ('disp_target',)

    def __init__(self, graph, disp_target):
        self.graph = graph
        self.disp_target = disp_target

    def marker_graph(self):
        return self.graph

    @property
    def target(self):
        return self.graph.value

if True: # pragma: no cover
    # Untested, experimental feature.
    class _StaticDispatchGraph(vg.PlainGraphNode, DispatchGraph):
        '''
        This is an untested stub, not yet complete, but a note on how to make
        static dispatch graphs. Basically a StaticDispatchGraph is a dispatch
        graph where the dispatch is done once, beforehand, and the resulting
        functions are cached in the graph. An experiment on a list of ~3600
        SchemaObjs suggests that precomputing the dispatch like this makes
        undictify at least 30% faster. I suspect that for larger, more complex
        typegraphs the savings will be dramatically more.

        Also, it'd be even more awesome to inspect Wrappers and pre-dispatch on
        their internal types, too, so that .for_marker() would often not need to
        do any dispatch. Another option would be to have a structure that caches
        dispatches made, so that functions fetched via parent() and super()
        could be dispatched once and never hit again...
        '''
        __slots__ = ('value',)

        def _get_fn(self):
            return self.value[2]

    def _predispatch(target_and_disp):
        target, disp = target_and_disp[0], target_and_disp[1]
        return (target, disp, disp.dispatch(target))

    def _bake(dispgraph):
        return vg.map(dispgraph, _predispatch, cls=_StaticDispatchGraph)

# These two wrapper classes are the reason why we generally don't add new
# methods to GraphNode subclasses - helpful tools like ValueOverlay and
# EdgeRestriction need to be reimplemented as well.
class DispatchOverlay(vg.ValueOverlay, DispatchGraph):
    '''DispatchGraph that overlays another DispatchGraph with changed attrs.'''
    __slots__ = ()
    def __init__(self, graph, target=None, disp_target=None):
        if target is None:
            target = graph.target
        if disp_target is None:
            disp_target = graph.disp_target
        super(DispatchOverlay, self).__init__(graph, (target, disp_target))

    def marker_graph(self):
        return vg.ValueOverlay(self.graph.marker_graph(), self.target)


class DispatchRestriction(vg.EdgeRestriction, DispatchGraph):
    '''DispatchGraph that overlays another with restricted edges.'''
    __slots__ = ()
    def marker_graph(self):
        return vg.EdgeRestriction(self.graph.marker_graph(), self.edge_names)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name='travesty',
    version='0.1.1',
    license='BSD',
    author="Daniel Lepage",
    author_email="dplepage@gmail.com",
    packages=['travesty','travesty.cantrips', 'travesty.document'],
    long_description="""
=======================================
 Travesty: Graph Traversal Dispatchers
=======================================

Travesty is a collection of tools for doing function dispatch based on a
graph.

A lot of these tools are specifically aimed at doing function dispatch based
on a type graph for some object type.

See README.rst for more info.

""",
    url='https://github.com/dplepage/travesty',
    install_requires=[
        'vertigo'
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 2",
    ]
)


from . import Wrapper, validate, InvalidAggregator

class Validated(Wrapper):
    '''Wrapper that specifies additional validators for a marker.

    >>> from travesty import Int
    >>> from travesty.validators import IsOneOf
    >>> typegraph = Validated.wrap(Int(), [IsOneOf([1,2,3])])
    >>> validate(typegraph, 2)
    >>> validate(typegraph, 15)
    Traceback (most recent call last):
        ...
    Invalid: invalid_choice

    Note that the type's validator is run first, and extra validators aren't run
    unless it passes. So e.g. if we pass a non-int in with our typegraph, we get
    a type_error instead of an invalid_choice.

    >>> validate(typegraph, "hello")
    Traceback (most recent call last):
        ...
    Invalid: type_error
    '''
    def __init__(self, marker, vdators=()):
        super(Validated, self).__init__(marker)
        self.vdators = tuple(vdators)

    @classmethod
    def wrap(cls, marker, vdators=()):
        return super(Validated, cls).wrap(marker, vdators=vdators)

@validate.when(Validated)
def validate_validated(dispgraph, value, **kwargs):
    validated = dispgraph.marker
    # If core validation fails, don't bother with higher-level validation
    dispgraph.for_marker(validated.marker)(value, **kwargs)
    # Now run each extra validator in turn
    fail_early = kwargs.get("dfy_fail_early", False)
    error_agg = InvalidAggregator(autoraise = fail_early)
    for vdator in validated.vdators:
        with error_agg.checking():
            dispgraph.for_marker(vdator)(value, **kwargs)
    error_agg.raise_if_any()



import sys
if sys.version < '3': # pragma: no cover
    unicode_type = unicode
    bytes_type = str
else: # pragma: no cover
    bytes_type = bytes
    unicode_type = str
    basestring = str

import vertigo as vg

from .invalid import Invalid, InvalidAggregator
from .base import Marker, traverse, validate, dictify, undictify, to_typegraph
from .schema import Schema

class SchemaMapping(Schema):
    '''Marker for structured dicts.

    The children of this node in the typegraph indicate the types of the keyed
    attributes to traverse.

    >>> from collections import OrderedDict
    >>> from datetime import date
    >>> from . import Int, Date
    >>> G = vg.PlainGraphNode

    >>> bday = date(1985, 9, 16)
    >>> schema = G(SchemaMapping(), [("age",G(Int())), ("birthday",G(Date()))])
    >>> def mk(age, birthday):
    ...     return OrderedDict([('age', age), ('birthday', birthday)])
    >>> print(vg.ascii_tree(traverse(schema, mk(age=27, birthday=bday))))
    root: OrderedDict([('age', 27), ('birthday', datetime.date(1985, 9, 16))])
      +--age: 27
      +--birthday: datetime.date(1985, 9, 16)

    >>> validate(schema, mk(age=27, birthday=bday))
    >>> validate(schema, mk(age=27.1, birthday="not a date"))
    Traceback (most recent call last):
    ...
    Invalid: age: [type_error], birthday: [type_error]

    >>> cstruct = dictify(schema, mk(age=27, birthday=bday))
    >>> cstruct == mk(age=27, birthday=u'1985-09-16')
    True
    >>> undictify(schema, cstruct) == mk(age=27, birthday=bday)
    True

    The 'extra_field_policy' flag should be 'save', 'discard', or 'error'. This
    controls what happens on undictifiance when unexpected fields are present.

    'save': un/dictify preserves extra keys; validate ignores them

    >>> schema = G.build(dict(_self=SchemaMapping('save'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27, 'color': 'blue'}
    True
    >>> validate(schema, v)
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27, 'color': 'blue'}
    True

    'discard': un/dictify discards extra keys; validate complains about them

    >>> schema = G.build(dict(_self=SchemaMapping('discard'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True
    >>> validate(schema, v)
    >>> validate(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True

    'error': undictify and validate complain; dictify discards

    >>> schema = G.build(dict(_self=SchemaMapping('error'), age=Int()))
    >>> v = undictify(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> validate(schema, {'age':27, 'color':'blue'})
    Traceback (most recent call last):
    ...
    Invalid: unexpected_fields - {'keys': set(['color'])}
    >>> v = dictify(schema, {'age':27, 'color':'blue'})
    >>> v == {'age': 27}
    True

    >>> validate(schema, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'int'>
    '''
    def __init__(self, extra_field_policy="discard"):
        super(SchemaMapping, self).__init__()
        self.extra_field_policy=extra_field_policy

@validate.when(SchemaMapping)
def validate_mapping(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid("type_error", "Expected dict, got {}".format(type(value)))
    marker = dispgraph.marker
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    with error_agg.checking():
        dispgraph.super(SchemaMapping)(value, **kwargs)
    if marker.extra_field_policy in ['error', 'discard']:
        extra_keys = set(value.keys()) - set(dispgraph.key_iter())
        if extra_keys:
            error_agg.own_error(Invalid('unexpected_fields', keys=extra_keys))
    error_agg.raise_if_any()

@dictify.when(SchemaMapping)
def dictify_mapping(dispgraph, value, **kwargs):
    marker = dispgraph.marker
    result = dispgraph.super(SchemaMapping)(value, **kwargs)
    if marker.extra_field_policy == 'save':
        for key in set(value.keys()) - set(dispgraph.key_iter()):
            result[key] = value[key]
    return result


@undictify.when(SchemaMapping)
def undictify_mapping(dispgraph, value, **kwargs):
    marker = dispgraph.marker
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    with error_agg.checking():
        result = dispgraph.super(SchemaMapping)(value, **kwargs)
    extra_keys = set(value.keys()) - set(dispgraph.key_iter())
    if marker.extra_field_policy == 'error' and extra_keys:
        error_agg.own_error(Invalid('unexpected_fields', keys=extra_keys))
    elif marker.extra_field_policy == 'save':
        for key in extra_keys:
            result[key] = value[key]
    error_agg.raise_if_any()
    return result


class StrMapping(Marker):
    '''Marker for dicts with string keys and homogenous values.

    The sub child in the typegraph determines the types of the mapping's values.

    >>> from collections import OrderedDict
    >>> from vertigo import PlainGraphNode as G
    >>> from . import Int, Date, List
    >>> from datetime import date
    >>> StringToNumList = StrMapping().of(List().of(Int()))
    >>> example = OrderedDict([("foo",[1,2,3,4]), ("bar",[6,12,-4])])
    >>> print(vg.ascii_tree(traverse(StringToNumList, example), sort=True))
    root: OrderedDict([('foo', [1, 2, 3, 4]), ('bar', [6, 12, -4])])
      +--bar: [6, 12, -4]
      |  +--0: 6
      |  +--1: 12
      |  +--2: -4
      +--foo: [1, 2, 3, 4]
         +--0: 1
         +--1: 2
         +--2: 3
         +--3: 4
    >>> cstruct = dictify(StringToNumList, example)
    >>> cstruct == {'foo':[1,2,3,4], 'bar':[6,12,-4]}
    True
    >>> undictify(StringToNumList, cstruct) == example
    True
    >>> undictify(StringToNumList, None)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'Nonetype'>
    >>> undictify(StringToNumList, {12:[1,2,3]})
    Traceback (most recent call last):
    ...
    Invalid: value_error/bad_keys - Bad keys - {'keys': [12]}
    >>> validate(StringToNumList, example)
    >>> validate(StringToNumList, {'foo':[1,"hi",3]})
    Traceback (most recent call last):
    ...
    Invalid: foo: [1: [type_error]]
    >>> validate(StringToNumList, {12:[1,2,3]})
    Traceback (most recent call last):
    ...
    Invalid: value_error/bad_keys - Bad keys - {'keys': [12]}
    >>> validate(StringToNumList, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'int'>

    >>> try:
    ...     undictify(StringToNumList, OrderedDict([
    ...         ('hello', None),
    ...     ]))
    ... except Invalid as e:
    ...     print(vg.ascii_tree(e.as_graph(), sort=True))
    root: []
      +--hello: [SingleInvalid('type_error',)]
    '''
    def of(self, sub):
        return vg.PlainGraphNode(self, sub = to_typegraph(sub))

@traverse.when(StrMapping)
def traverse_strmap(dispgraph, value, zipgraph=None, **kwargs):
    edges = []
    valgraph = dispgraph['sub']
    valzip = zipgraph['sub'] if zipgraph else None
    for key, val in value.items():
        edges.append((key, valgraph(val, valzip, **kwargs)))
    v = value
    if zipgraph:
        v = (v, zipgraph.value)
    return vg.PlainGraphNode(v, edges)

@undictify.when(StrMapping)
def undictify_strmap(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid("type_error", "Expected dict, got {}".format(type(value)))
    # If fail_early is True, then gather all errors from this and its
    # children. Otherwise, just raise the first error we encounter.
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    data = {}
    bad_keys = []
    for key, val in value.items():
        if not isinstance(key, basestring):
            bad_keys.append(key)
            continue
        with error_agg.checking_sub(key):
            val = dispgraph['sub'](val, **kwargs)
        data[key] = val
    if bad_keys:
        with error_agg.checking():
            raise Invalid("value_error/bad_keys", "Bad keys", keys=bad_keys)
    error_agg.raise_if_any()
    return data

@dictify.when(StrMapping)
def dictify_strmap(dispgraph, value, **kwargs):
    sub = dispgraph['sub']
    return {key:sub(val, **kwargs) for (key, val) in value.items()}

@validate.when(StrMapping)
def validate_strmap(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid("type_error", "Expected dict, got {}".format(type(value)))
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    bad_keys = []
    for key, val in value.items():
        if not isinstance(key, basestring):
            bad_keys.append(key)
            continue
        with error_agg.checking_sub(key):
            dispgraph['sub'](val, **kwargs)
    if bad_keys:
        with error_agg.checking():
            raise Invalid("value_error/bad_keys", "Bad keys", keys=bad_keys)
    error_agg.raise_if_any()

class UniMapping(Marker):
    '''Marker for dicts with homogenous keys and homogenous values.

    The key and val children in the typegraph determine the types of the
    mapping's keys and values, respectively.

    >>> from collections import OrderedDict
    >>> from vertigo import PlainGraphNode as G
    >>> from . import Int, Date, List
    >>> from datetime import date
    >>> DateToNumList = G.build(dict(
    ...     _self=UniMapping(),
    ...     key=Date(),
    ...     val=dict(_self=List(), sub=Int()))
    ... )
    >>> d1, d2 = date(1985, 9, 16), date(1980, 3, 17)
    >>> example = OrderedDict([(d1,[1,2,3,4]), (d2,[6,12,-4])])
    >>> print(vg.ascii_tree(traverse(DateToNumList, example), sort=True))
    root: OrderedDict([(datetime.date(1985, 9, 16), [1, 2, 3, 4]), (datetime.date(1980, 3, 17), [6, 12, -4])])
      +--key_0: datetime.date(1985, 9, 16)
      +--key_1: datetime.date(1980, 3, 17)
      +--value_0: [1, 2, 3, 4]
      |  +--0: 1
      |  +--1: 2
      |  +--2: 3
      |  +--3: 4
      +--value_1: [6, 12, -4]
         +--0: 6
         +--1: 12
         +--2: -4
    >>> cstruct = dictify(DateToNumList, example)
    >>> cstruct == {u'1985-09-16':[1,2,3,4], '1980-03-17':[6,12,-4]}
    True
    >>> undictify(DateToNumList, cstruct) == example
    True
    >>> undictify(DateToNumList, None)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'Nonetype'>
    >>> validate(DateToNumList, example)
    >>> validate(DateToNumList, {"not a date":[1,2,3]})
    Traceback (most recent call last):
    ...
    Invalid: key_0: [type_error]
    >>> validate(DateToNumList, {d1:[1,"hi",3]})
    Traceback (most recent call last):
    ...
    Invalid: value_0: [1: [type_error]]
    >>> validate(DateToNumList, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error - Expected dict, got <type 'int'>

    >>> try:
    ...     undictify(DateToNumList, OrderedDict([
    ...         ('1985-09-16', [1, 'hi']),
    ...         ('hello', None),
    ...     ]))
    ... except Invalid as e:
    ...     print(vg.ascii_tree(e.as_graph(), sort=True))
    root: []
      +--key_1: [SingleInvalid('bad_format',)]
      +--value_1: [SingleInvalid('type_error',)]

    Caveat: Note that if the value passed in is not an OrderedDict, then there's
    no guarantee of the numerical order of the errors or of the traversal -
    '''
    def of(self, key, val):
        key, val = to_typegraph(key), to_typegraph(val)
        return vg.PlainGraphNode(self, key=key, val=val)

@traverse.when(UniMapping)
def traverse_unimap(dispgraph, value, zipgraph=None, **kwargs):
    edges = []
    keygraph = dispgraph['key']
    valgraph = dispgraph['val']
    keyzip = zipgraph['key'] if zipgraph else None
    valzip = zipgraph['val'] if zipgraph else None
    for i, (key, val) in enumerate(value.items()):
        edges.append(('key_{}'.format(i), keygraph(key, keyzip, **kwargs)))
        edges.append(('value_{}'.format(i), valgraph(val, valzip, **kwargs)))
    v = value
    if zipgraph:
        v = (v, zipgraph.value)
    return vg.PlainGraphNode(v, edges)


@undictify.when(UniMapping)
def undictify_unimap(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid("type_error", "Expected dict, got {}".format(type(value)))
    # If fail_early is True, then gather all errors from this and its
    # children. Otherwise, just raise the first error we encounter.
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    data = {}
    for i, (key, val) in enumerate(value.items()):
        with error_agg.checking_sub('key_{}'.format(i)):
            key = dispgraph['key'](key, **kwargs)
        with error_agg.checking_sub('value_{}'.format(i)):
            val = dispgraph['val'](val, **kwargs)
        data[key] = val
    error_agg.raise_if_any()
    return data

@dictify.when(UniMapping)
def dictify_unimap(dispgraph, value, **kwargs):
    kdfy = lambda x: dispgraph['key'](x, **kwargs)
    vdfy = lambda x: dispgraph['val'](x, **kwargs)
    return {kdfy(key):vdfy(val) for (key, val) in value.items()}

@validate.when(UniMapping)
def validate_unimap(dispgraph, value, **kwargs):
    if not isinstance(value, dict):
        raise Invalid("type_error", "Expected dict, got {}".format(type(value)))
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    for i, (key, val) in enumerate(value.items()):
        with error_agg.checking_sub('key_{}'.format(i)):
            dispgraph['key'](key, **kwargs)
        with error_agg.checking_sub('value_{}'.format(i)):
            dispgraph['val'](val, **kwargs)
    error_agg.raise_if_any()



if __name__ == '__main__': # pragma: no cover
    import doctest
    doctest.testmod()




'''

>>> from travesty import String, dictify, undictify, traverse, validate, Optional
>>> import vertigo as vg
>>> class Node(Document):
...     # field_types is a fn because python lacks recursive class definitions
...     field_types = lambda cls: dict(
...         name=String(),
...         next=Optional.wrap(cls)
...     )
...     def __init__(self, uid, name, next=None):
...         super(Node, self).__init__(uid)
...         self.name, self.next = name, next
...     def _loaded_repr(self):
...         return "Node({!r}, {!r})".format(self.uid, self.name)
>>> Node._finalize_typegraph()

You can create two nodes that point to each other:

>>> n1 = Node(u"node1", u"First")
>>> n2 = Node(u"node2", u"Second", next=n1)
>>> n1.next = n2

Dictifying without passing in a storage pit behaves just like dictfying a any
other recursive SchemaObj, i.e. it overflows the stack:

>>> dictify(Node, n1)
Traceback (most recent call last):
    ...
RuntimeError: maximum recursion depth exceeded

However, if pass in a dictionary called doc_storage, then dictifying will only
return {'uid':uid}, and the doc_storage dict will be filled with the encountered
documents:

>>> store = {}
>>> dictify(Node, n1, doc_storage=store)
{'uid': u'node1'}
>>> n1_d = store['node1']
>>> n2_d = store['node2']
>>> n1_d == dict(uid='node1', next=dict(uid='node2'), name="First")
True
>>> n2_d == dict(uid='node2', next=dict(uid='node1'), name=u"Second")
True

Then you can undictify them with a DocSet:

>>> loader = DocSet()
>>> node1 = loader.load(Node, n1_d)
>>> node1.name
u'First'

After loading just node1, its next will be an unloaded reference to node2:

>>> node1.next.uid
u'node2'
>>> node1.next.loaded
False
>>> node1.next.name
Traceback (most recent call last):
    ...
UnloadedDocumentException: <Unloaded Node: node2>

But after loading node2, node1.next will be loaded (and will be node2):

>>> node2 = loader.load(Node, n2_d)
>>> node1.next.loaded
True
>>> node1.next.name
u'Second'
>>> node1.next is node2
True
>>> node1.next.next is node1
True

validate and traverse will follow docrefs; be careful not to use them on
recursive structures.

>>> try:
...     validate(Node, node1)
... except RuntimeError as e:
...     assert str(e).startswith("maximum recursion depth exceeded")
... else:
...     raise Exception("That should have failed.")
>>> node2.next = None
>>> validate(Node, node1)
>>> print(vg.ascii_tree(traverse(Node, node1), sort=True))
root: Node(u'node1', u'First')
  +--name: u'First'
  +--next: Node(u'node2', u'Second')
  |  +--name: u'Second'
  |  +--next: None
  |  +--uid: u'node2'
  +--uid: u'node1'



You can also create a DocSet from existing documents, but the DocSet will
NOT automatically search for other referenced documents:

>>> docset = DocSet([node1])
>>> docset[Node, 'node1'] is node1
True
>>> (Node, 'node2') in docset
False
>>> docset.get(Node, 'node2') is None
True

You can add other documents explicitly by calling add:

>>> docset.add(node2)
>>> docset[Node, 'node2'] is node2
True

This will fail if a document with that id is already in the set:

>>> docset.add(node2)
Traceback (most recent call last):
    ...
ValueError: Duplicate uid node2 for type <class 'travesty.document.Node'>

Finally, the explicit DocSet.create(type, uid) function creates an unloaded
document, or fails if that document already exists:

>>> docset.create(Node, 'node3')
<Unloaded Node: node3>
>>> docset.create(Node, 'node1')
Traceback (most recent call last):
    ...
ValueError: Duplicate uid node1 for type <class 'travesty.document.Node'>
'''


from .document import Document, UnloadedDocumentException, DoubleLoadException
from .docset import DocSet

__all__ = ['Document', 'DocSet', 'UnloadedDocumentException',
    'DoubleLoadException']




import vertigo as vg

from .base import Marker, traverse, validate, dictify, undictify
from .base import to_typegraph
from .invalid import Invalid, InvalidAggregator

class List(Marker):
    '''Marker for homogenous lists.

    The 'sub' child in the typegraph indicates the marker for all items in the
    list.

    For example:

    >>> from travesty import Int, Optional
    >>> l = List().of(Int())

    >>> print(vg.ascii_tree(traverse(l, [1, 2, 3]), sort=True))
    root: [1, 2, 3]
      +--0: 1
      +--1: 2
      +--2: 3

    >>> undictify(l, [1, 4, 5])
    [1, 4, 5]
    >>> dictify(l, [1, 2, 3])
    [1, 2, 3]

    >>> validate(l, [1, 2, 3])
    >>> try:
    ...     validate(l, [1, None, 'hi'])
    ... except Invalid as e:
    ...     print(vg.ascii_tree(e.as_graph(), sort=True))
    root: []
      +--1: [SingleInvalid('type_error',)]
      +--2: [SingleInvalid('type_error',)]
    >>> l = List().of(Optional.wrap(Int()))
    >>> validate(l, [1, None, 'hi'])
    Traceback (most recent call last):
    ...
    Invalid: 2: [type_error]
    >>> validate(l, 12)
    Traceback (most recent call last):
    ...
    Invalid: type_error
    '''
    def of(self, sub):
        return vg.PlainGraphNode(self, sub=to_typegraph(sub))

@traverse.when(List)
def traverse_list(dispgraph, value, zipgraph=None, **kwargs):
    edges = []
    subzip = zipgraph['sub'] if zipgraph else None
    for i,val in enumerate(value):
        key = str(i)
        node = dispgraph['sub'](val, zipgraph=subzip, **kwargs)
        edges.append((key,node))
    v = value
    if zipgraph:
        v = (v, zipgraph.value)
    return vg.PlainGraphNode(v, edges)

@validate.when(List)
def validate_list(dispgraph, value, **kwargs):
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    try:
        iterator = iter(value)
    except TypeError: # item isn't iterable!
        raise Invalid("type_error", "Value is not iterable")
    for i, item in enumerate(iterator):
        with error_agg.checking_sub(str(i)):
            dispgraph['sub'](item, **kwargs)
    error_agg.raise_if_any()

@undictify.when(List)
def undictify_list(dispgraph, value, **kwargs):
    # If _full_errors is True, then gather all errors from this and its
    # children. Otherwise, just raise the first error we encounter.
    error_agg = InvalidAggregator(autoraise = kwargs.get('fail_early', False))
    try:
        iterator = iter(value)
    except TypeError: # item isn't iterable!
        raise Invalid("type_error", "Value is not iterable")
    result = []
    for i, item in enumerate(iterator):
        with error_agg.checking_sub(str(i)):
            result.append(dispgraph['sub'](item, **kwargs))
    error_agg.raise_if_any()
    return result

@dictify.when(List)
def dictify_list(dispgraph, value, **kwargs):
    return [dispgraph['sub'](x, **kwargs) for x in value]

if __name__ == '__main__': # pragma: no cover
    import doctest
    doctest.testmod()
